# -*- coding: utf-8 -*-
"""
Created on Tue Oct  3 08:04:49 2023

@author: mfratki
"""

import pandas as pd
from mpcaHydro import pywisk
#from hspf_tools.orm.models import Station
import time


'''
Q - Discharge - cfs/s

Constituents
WT - Water Temperature - degrees C
WL - Water Level - ft
OP - Orthophosphate - mg/L
DO - Dissolved Oxygen - mg/L
TP - Total Phosphorus - mg/L
TSS - Total Suspended Solids - mg/L
N - Nitrogen - mg/L
TKN - Total Kjeldahl Nitrogen - mg/L
CHL - Chlorophyll - ug/L
BOD - Biochemical Oxygen Demand - mg/L


load is a derived measurement

'''


PARAMETERTYPE_MAP ={'11522': 'TP',
                    '11531': 'TP',
                    '11532': 'TSS',
                    '11523': 'TSS',
                    '11526': 'N',
                    '11519': 'N',
                    '11520': 'OP',
                    '11528': 'OP',
                    '11530': 'TKN',
                    '11521': 'TKN',
                    '11500' : 'Q',
                    '11504': 'WT',
                    '11533': 'DO'}
#                    '11507':'WL'}
#STATIONPARAMETER_NOS = ['262*','450*','451*','863*','866*','5034' ,'5035','5005', '5004','5014' ,'5015','5024'  ,'5025','5044' ,'5045']
STATIONPARAMETER_NOS = ['262*','450*','451*','863*','866*']

DATA_CODES = [1,3,10,12,
                               15,
                               20,
                               29,
                               30,
                               31,
                               32,
                               34,
                               45,
                               46,
                               47,
                               48,
                               49]



# TS_NAME_SELECTOR = {'Q':{'Internal':['20.Day.Mean.Archive','15.Rated'],
#                          'External': ['20.Day.Mean','08.Provisional.Edited']},
#                     'WT':{'Internal':['20.Day.Mean','09.Archive'],
#                           'External': ['20.Day.Mean','08.Provisional.Edited']},
#                     'TSS':{'Internal':['20.Day.Mean','09.Archive'],
#                           'External': ['20.Day.Mean','08.Provisional.Edited']},
#                     'N':{'Internal':['20.Day.Mean','09.Archive'],
#                           'External': ['20.Day.Mean','08.Provisional.Edited']},
#                     'TKN':{'Internal':['20.Day.Mean','09.Archive'],
#                           'External': ['20.Day.Mean','08.Provisional.Edited']},
#                     'TP':{'Internal':['20.Day.Mean','09.Archive'],
#                           'External': ['20.Day.Mean','08.Provisional.Edited']},
#                     'OP':{'Internal':['20.Day.Mean','09.Archive'],
#                           'External': ['20.Day.Mean','08.Provisional.Edited']},
#                     'DO':{'Internal':['20.Day.Mean','09.Archive'],
#                           'External': ['20.Day.Mean','08.Provisional.Edited']}}


TS_NAME_SELECTOR = {'Q':{'Internal':{'daily':'20.Day.Mean.Archive',
                                     'unit': '15.Rated'},
                         'External': {'daily': '20.Day.Mean',
                                      'unit': '08.Provisional.Edited'}},
                    'WT':{'Internal':{'daily':'20.Day.Mean',
                                      'unit': '09.Archive'},
                          'External': {'daily': '20.Day.Mean',
                                       'unit': '08.Provisional.Edited'}},
                    'TSS':{'Internal':{'daily':'20.Day.Mean',
                                      'unit': '09.Archive'},
                          'External': {'daily': '20.Day.Mean',
                                       'unit': '08.Provisional.Edited'}},                   
                    'N':{'Internal':{'daily':'20.Day.Mean',
                                      'unit': '09.Archive'},
                          'External': {'daily': '20.Day.Mean',
                                       'unit': '08.Provisional.Edited'}},                    
                    'TKN':{'Internal':{'daily':'20.Day.Mean',
                                      'unit': '09.Archive'},
                          'External': {'daily': '20.Day.Mean',
                                       'unit': '08.Provisional.Edited'}},                    
                    'TP':{'Internal':{'daily':'20.Day.Mean',
                                      'unit': '09.Archive'},
                          'External': {'daily': '20.Day.Mean',
                                       'unit': '08.Provisional.Edited'}},                    
                    'OP':{'Internal':{'daily':'20.Day.Mean',
                                      'unit': '09.Archive'},
                          'External': {'daily': '20.Day.Mean',
                                       'unit': '08.Provisional.Edited'}},                    
                    'DO':{'Internal':{'daily':'20.Day.Mean',
                                      'unit': '09.Archive'},
                          'External': {'daily': '20.Day.Mean',
                                       'unit': '08.Provisional.Edited'}},
                    'TRB':{'Internal':{'daily':'20.Day.Mean',
                                    'unit': '09.Archive'},
                        'External': {'daily': '20.Day.Mean',
                                    'unit': '08.Provisional.Edited'}},
                }



CONSTITUENT_NAME_NO = {'Q'  :['262*'],#,'263'],
                       'WT' :['450*', '451*'], # '450.42','451.42'],
                       'OP' :['863*'],
                       'DO' :['866*'],
                       'TRB': ['811*'],
                       'TDS': ['2175*'],
                       'TP' :None,
                       'TSS':None,
                       'N'  :['341*'],
                       'TKN':None}

CONSTITUENT_NAME_NO_WPLMN = {'Q'  :['262*'],#,'263'],
                       'WT' :['450*', '451*'], # '450.42','451.42'],
                       'OP' :['863*','5034' ,'5035'],
                       'DO' :['866*'],
                       'TP' :['5005'  ,'5004'],
                       'TSS':['5014' ,'5015'],
                       'N'  :['5024'  ,'5025'],
                       'TKN':['5044' ,'5045']}

VALID_CONSTITUENTS = ['Q','WT','OP','DO','TP','TSS','N','TKN','TRB']

# def _info(station_nos):
#     station_info = info(station_nos)
#     if station_info.empty:
#         return Station(station_nos,
#                        'wiski',
#                        station_name = '',
#                        station_type = 'River')
#     return Station(station_info.iloc[0]['station_no'],
#                    'wiski',
#                    station_name = station_info.iloc[0]['station_name'],
#                    latitude = station_info.iloc[0]['station_latitude'],
#                    longitude = station_info.iloc[0]['station_longitude'],
#                    station_type = 'River')


def extract(station_nos, constituent, dbpath, start_year = 1996, end_year = 2030, wplmn = False):
    '''
    given a list of station_nos, download all data relevent to HSPF from MPCA WISKI and store in a duckdb database
    
    1. Find relevent timeseries ids for each constituent
    2. Download data for each timeseries id
    3. Store data in duckdb database
    
    '''
    #1. Find relevent timeseries ids for each constituent
    if station_nos[0] == 'E':
        ts_names = TS_NAME_SELECTOR[constituent]['External']
    else:
        ts_names =TS_NAME_SELECTOR[constituent]['Internal']
    
    if wplmn:
        constituent_nos = CONSTITUENT_NAME_NO_WPLMN[constituent]
    else:
        constituent_nos = CONSTITUENT_NAME_NO[constituent]
        
    ts_ids = pywisk.get_ts_ids(station_nos = station_nos,
                        stationparameter_no = constituent_nos,
                        ts_name = ts_names['unit'])
    
    jsons = []
    for ts_id in ts_ids:
        jsons.append(download_chunk(ts_id,start_year,end_year,as_json = True))
        time.sleep(.1)



    
    # Connect to DuckDB (in-memory database)
    con = duckdb.connect(database=':memory:')

    # Register the Python list of dictionaries as a virtual table
    # DuckDB can automatically infer the schema from this list.
    con.register("my_json_table", json_data)
    return jsons


    # with duckdb.connect(db_path) as con:
    #     con.execute("DROP TABLE IF EXISTS observations")
    #     datafiles = folderpath.joinpath('*.csv').as_posix()
    #     query = '''
    #     CREATE TABLE observations AS SELECT * 
    #     FROM
    #     read_csv_auto(?,
    #                     union_by_name = true);
        
    #     '''
    #     con.execute(query,[datafiles])

    
    # con = duckdb.connect(database=db_path))
    # print('Downloading Timeseries Data')
    # df = pd.concat([_download(constituent,station_nos,start_year,end_year,raw = True, wplmn = False) for constituent in VALID_CONSTITUENTS])
    # df.to_csv(filepath,index = False)
    # print('Timeseries Data Downloaded!')

    


def info(station_nos):
    ts_ids = pywisk.get_ts_ids(station_nos = station_nos,
                        stationparameter_no = STATIONPARAMETER_NOS)
    ts_ids = ts_ids.drop_duplicates(subset = 'parametertype_name')
    ts_ids['constituent'] = ts_ids['parametertype_id'].map(PARAMETERTYPE_MAP)
    ts_ids = ts_ids.loc[:,['station_no','station_name','stn_HUC12','stn_EQuIS_ID','stn_AUID','constituent','parametertype_name','station_latitude','station_longitude']]
    return ts_ids    

def download(station_nos,start_year = 1996, end_year = 2030, raw = False,wplmn = False):

    
    print('Downloading Timeseries Data')
    df = pd.concat([_download(constituent,station_nos,start_year,end_year,raw,wplmn) for constituent in VALID_CONSTITUENTS])
    
    station_metadata = pywisk.get_stations(station_no = station_nos,returnfields = ['stationgroup_id'])
    if any(station_metadata['stationgroup_id'].isin(['1319204'])):
        df['wplmn_flag'] = 1
    else:
        df['wplmn_flag'] = 0
    print('Done!')
    return df

def transform(data):

    data['datetime'] = pd.to_datetime(data.loc[:,'Timestamp'])
    data = data.loc[data['Quality Code'].isin(DATA_CODES)].copy()
    data.rename(columns = {'Value': 'value',
                           'stationparameter_name': 'variable',
                           'station_name': 'station_name',
                           'station_no': 'station_id',
                           'Quality Code': 'quality_id',
                           'ts_unitsymbol':'unit'},inplace = True)
    
    #if data['ts_name'].str

    
    data['constituent'] = data['parametertype_id'].map(PARAMETERTYPE_MAP)
    
    data = data.loc[:,['datetime','value','variable','unit','station_id','station_name','constituent','data_format','quality_id','interval_minutes']]
    
    
    data.loc[data['unit'] == 'kg','value'] = data.loc[data['unit'] == 'kg','value'].apply(lambda x: (x*2.20462))
    data.replace({'unit':'kg'},'lb',inplace=True)
    data.replace({'unit':'ft³/s'},'cfs',inplace=True)
    data.loc[:,'unit'] = data['unit'].str.lower()
    data.replace({'unit':'°f'},'degF',inplace = True)
    data['data_type'] = 'continuous'
    data['station_origin'] = 'wiski'
    data.set_index('datetime',drop=True,inplace=True)
    data.index = data.index.tz_convert('UTC-06:00')
    
    
    data.index = data.index.round('h').round('h')
    data = data.reset_index()
    data = data.groupby(['datetime','variable','unit','station_id','station_name','constituent','interval_minutes','data_format','data_type','station_origin']).mean()
    data = data.reset_index()
    data = data.set_index('datetime')
    
    
    
    
    return data

def load(data,file_path):
    data.to_csv(file_path)



def _download(constituent,station_nos,start_year = 1996,end_year = 2030, raw = False,wplmn = False):

    if station_nos[0] == 'E':
        ts_names = TS_NAME_SELECTOR[constituent]['External']
    else:
        ts_names =TS_NAME_SELECTOR[constituent]['Internal']
    
    if wplmn:
        constituent_nos = CONSTITUENT_NAME_NO_WPLMN[constituent]
    else:
        constituent_nos = CONSTITUENT_NAME_NO[constituent]
        
    ts_ids = pywisk.get_ts_ids(station_nos = station_nos,
                        stationparameter_no = constituent_nos,
                        ts_name = ts_names['unit'])
    
    interval_minutes = 60
    if ts_ids.empty:
        ts_ids = pywisk.get_ts_ids(station_nos = station_nos,
                            stationparameter_no = constituent_nos,
                            ts_name = ts_names['daily'])
        interval_minutes = 1440 
        if ts_ids.empty:
            return pd.DataFrame()    
    
    df = convert_to_df(ts_ids['ts_id'],start_year,end_year)

    if df.empty:
        return pd.DataFrame()    
    
    if constituent == 'WT':
        df.loc[:,'ts_unitsymbol'] = df['ts_unitsymbol'].str.lower()
        df.loc[df['ts_unitsymbol'] == '°c','Value'] = df.loc[df['ts_unitsymbol'] == '°c','Value'].apply(lambda x: (x*9/5)+32)
        df.replace({'ts_unitsymbol':'°c'},'°f',inplace=True)
        
    df['data_format'] = 'instantaneous'
    df['interval_minutes'] = interval_minutes
    if not raw: df = transform(df)
    return df


def download_chunk(ts_id,start_year = 1996,end_year = 2030, interval = 5, as_json = False):
    frames = [pd.DataFrame()]

    for start in range(start_year,end_year,interval):
        end = int(start + interval-1)
        if end > end_year:
            end = end_year
        df = pywisk.get_ts(ts_id,start_date = f'{start}-01-01',end_date = f'{end}-12-31',as_json = as_json)
        if not df.empty: frames.append(df)
        df.index = pd.to_datetime(df['Timestamp'])
        time.sleep(.1)   
    return pd.concat(frames)

def convert_to_df(ts_ids,start_year = 1996,end_year = 2030):
    dfs = []
    for ts_id in ts_ids:
        dfs.append(download_chunk(ts_id,start_year,end_year))
        time.sleep(.1)
    df =  pd.concat(dfs)
    return df



def discharge(station_nos,start_year = 1996,end_year = 2030,raw=False):
    return _download('Q',station_nos,start_year,end_year,raw)


def temperature(station_nos,start_year = 1996,end_year = 2030,raw=False):
    return _download('WT',station_nos,start_year,end_year,raw)


def orthophosphate(station_nos,start_year = 1996,end_year = 2030,raw=False):
    return _download('OP',station_nos,start_year,end_year,raw)

def dissolved_oxygen(station_nos,start_year = 1996,end_year = 2030,raw=False):
    return _download('DO',station_nos,start_year,end_year,raw)

def nitrogen(station_nos,start_year = 1996,end_year = 2030,raw=False):
    return _download('N',station_nos,start_year,end_year,raw)

def total_suspended_solids(station_nos,start_year = 1996,end_year = 2030,raw=False):
    return _download('TSS',station_nos,start_year,end_year,raw)

def total_phosphorous(station_nos,start_year = 1996,end_year = 2030,raw=False):
    return _download('TP',station_nos,start_year,end_year,raw)

def tkn(station_nos,start_year = 1996,end_year = 2030,raw=False):
    return _download('TKN',station_nos,start_year,end_year,raw)










# def discharge(station_nos,start_year = 1996,end_year = 2030,raw=False):
#     if station_nos[0] == 'E':
#         ts_names = ['08.Provisional.Edited']
#     else:
#         ts_names = ['15.Rated']
        
#     ts_ids = pywisk.get_ts_ids(station_nos = station_nos,
#                         stationparameter_no = ['262*'],
#                         ts_name = ts_names)
    
#     if ts_ids.empty:
#         return pd.DataFrame()

    
#     df = convert_to_df(ts_ids['ts_id'],start_year,end_year)

#     if df.empty:
#         df = pd.DataFrame()
#     else:
#         df['data_format'] = 'instantaneous'
#         if not raw: df = transform(df)
#     return df



# def temperature(station_nos,start_year = 1996,end_year = 2030,raw=False):
#     if station_nos[0] == 'E':
#         ts_names = ['08.Provisional.Edited']
#     else:
#         ts_names = ['09.Archive']
  
    
#     ts_ids = pywisk.get_ts_ids(station_nos = station_nos,
#                       stationparameter_no = ['450*','450.42*','451*','451.42*'],
#                       ts_name = ts_names)
        
#     if ts_ids.empty:
#         return pd.DataFrame()
    
           
#     df = convert_to_df(ts_ids['ts_id'],start_year,end_year)
#     if df.empty:
#         df = pd.DataFrame()
#     else:
#         #df.loc[df['ts_unitsymbol'] == '°f','Value'] = df.loc[df['ts_unitsymbol'] == '°f','Value'].apply(lambda x: (x - 32)*5/9)
#         df.loc[df['ts_unitsymbol'] == '°c','Value'] = df.loc[df['ts_unitsymbol'] == '°c','Value'].apply(lambda x: (x*9/5)+32)

#         df['ts_unitsymbol'].replace('°c','°f',inplace=True)
#         df['data_format'] = 'instantaneous'
#         if not raw: df = transform(df)

#     return df

# def dissolved_oxygen(station_nos,start_year = 1996,end_year = 2030,raw=False):
#     if station_nos[0] == 'E':
#         ts_names = ['08.Provisional.Edited']
#     else:
#         ts_names = ['09.Archive']
  
#     ts_ids = pywisk.get_ts_ids(station_nos = station_nos,
#                       stationparameter_no = ['866*'],
#                       ts_name = ts_names)
        
#     if ts_ids.empty:
#         return pd.DataFrame()
#     df = convert_to_df(ts_ids['ts_id'],start_year,end_year)
    
#     if df.empty:
#         df = pd.DataFrame()
#     else:
#         df['data_format'] = 'instantaneous'
#         if not raw: df = transform(df)
#     return df


# def orthophosphate(station_nos,start_year = 1996,end_year = 2030,raw=False):
#     ts_ids = pywisk.get_ts_ids(station_nos = station_nos,
#                       stationparameter_no = ['5034','5035'],
#                       ts_name = ['20.Day.Mean'])
        
#     if ts_ids.empty:
#         return pd.DataFrame()
#     df = convert_to_df(ts_ids['ts_id'],start_year,end_year)

    
#     if df.empty:
#         df = pd.DataFrame()
#     else:
#         df['data_format'] = 'aggregated'
#         if not raw: df = transform(df)

#     return df

# def total_phosphorous(station_nos,start_year = 1996,end_year = 2030,raw=False):
  
#     ts_ids = pywisk.get_ts_ids(station_nos = station_nos,
#                       stationparameter_no = ['5004','5005'],
#                       ts_name = ['20.Day.Mean'])    
#     if ts_ids.empty:
#         return pd.DataFrame()
#     df = convert_to_df(ts_ids['ts_id'],start_year,end_year)
    
#     if df.empty:
#         df = pd.DataFrame()
#     else:
#         df['data_format'] = 'aggregated'
#         if not raw: df = transform(df)

#     return df


# def total_suspended_solids(station_nos,start_year = 1996,end_year = 2030,raw=False):
  
#     ts_ids = pywisk.get_ts_ids(station_nos = station_nos,
#                       stationparameter_no = ['5014','5015'],
#                       ts_name = ['20.Day.Mean'])
#     if ts_ids.empty:
#         return pd.DataFrame()
#     df = convert_to_df(ts_ids['ts_id'],start_year,end_year)
    
#     if df.empty:
#         df = pd.DataFrame()
#     else:
#         df['data_format'] = 'aggregated'
#         if not raw: df = transform(df)

#     return df


# def tkn(station_nos,start_year = 1996,end_year = 2030,raw=False):
  
#     ts_ids = pywisk.get_ts_ids(station_nos = station_nos,
#                       stationparameter_no = ['5044','5045'],
#                       ts_name = ['20.Day.Mean'])
        
#     if ts_ids.empty:
#         return pd.DataFrame()
#     df = convert_to_df(ts_ids['ts_id'],start_year,end_year)
   
#     if df.empty:
#         df = pd.DataFrame()
#     else:
#         df['data_format'] = 'aggregated'
#         if not raw: df = transform(df)

#     return df


# def nitrogen(station_nos,start_year = 1996,end_year = 2030,raw=False):
  
#     ts_ids = pywisk.get_ts_ids(station_nos = station_nos,
#                       stationparameter_no = ['5024','5025'],
#                       ts_name = ['20.Day.Mean'])
        
#     if ts_ids.empty:
#         return pd.DataFrame()
#     df = convert_to_df(ts_ids['ts_id'],start_year,end_year)
   
#     if df.empty:
#         df = pd.DataFrame()
#     else:
#         df['data_format'] = 'aggregated'
#         if not raw: df = transform(df)

#     return df



# station_nos = ['H57026001'] # Exception station. Flow data is from a different station upstream
# station_nos = ['H36061003']








# TS_NAMES = ['20.Day']
# wiski = pyWISK()
# #external vs internal station

# station_nos = 'E58059001'

# PARAMETERS_MAP={'5004':'TP Load',
#                 '5005':'TP Conc',
#                 '5014':'TSS Load',
#                 '5015':'TSS Conc',
#                 '5024':'N Load',
#                 '5025':'N Conc',
#                 '5034':'OP Load',
#                 '5035':'OP Conc',
#                 '5044':'TKN Load',
#                 '5045':'TKN Conc',
#                 '262*' :'Flow'}




# PARAMETERS = ['262*',
# '450*',
# '451*',
# '341*',
# '855*',
# '866*',]


# '501*',
# '502*',
# '503*',
# '504*']

# PARAMETER_NO_MAP = {'262':'Q',
#                  '263':'Q',
#                  '450':'WT',
#                  '451':'WT',
#                  '450.42':'WT',
#                  '451.42':'WT',
#                  '863':'OP',
#                  '865':'DO',
#                  '866':'DO',
#                  '867':'DO',
#                  '5005': 'TP',
#                  '5004': 'TP',
#                  '5014': 'TSS',
#                  '5015': 'TSS',
#                  '5024': 'N',
#                  '5025': 'N',
#                  '5034': 'OP',
#                  '5035': 'OP',
#                  '5044': 'TKN',
#                  '5045': 'TKN'}



# -*- coding: utf-8 -*-
"""
Created on Tue Oct 10 14:13:23 2023

@author: mfratki
"""

import pandas as pd
from pathlib import Path
#from hspf_tools.orm.models import Station
# import geopandas as gpd

EQUIS_PARAMETER_XREF = pd.read_csv(Path(__file__).parent/'data/EQUIS_PARAMETER_XREF.csv')


CONSTITUENT_MAP = {i[0]:i[1] for i in EQUIS_PARAMETER_XREF[['PARAMETER','constituent']].values}
  

# station_no  = 	'S010-822'
# data = download(station_no)
# data = transform(data) 


# def download(station_nos):
#     df = pd.concat([_download(station_no) for station_no in station_nos])
#     return df
import requests

def _download(station_no):
    # Replace {station_no} in the URL with the actual station number
    url = f"https://services.pca.state.mn.us/api/v1/surfacewater/monitoring-stations/results?stationId={station_no}&format=json"
    
    try:
        # Send a GET request to the URL
        response = requests.get(url)
        response.raise_for_status()  # Raise exception for HTTP errors
        # Parse the JSON data
        if response.json()['recordCount'] == 0:
            return pd.DataFrame(columns = response.json()['column_names'])
        else:
            return pd.DataFrame(response.json()['data'])
    
    except requests.exceptions.RequestException as e:
        print(f"An error occurred: {e}")
        return None



def download(station_no):
    #df = pd.read_csv(f'https://services.pca.state.mn.us/api/v1/surfacewater/monitoring-stations/results?stationId={station_no}&format=csv')
    df = _download(station_no)
    if df.empty:
        return df
    else:
        df['station_id'] = station_no
        return transform(df)

def info(station_no):
    #df = pd.read_csv(f'https://services.pca.state.mn.us/api/v1/surfacewater/monitoring-stations/results?stationId={station_no}&format=csv')
    df = _download(station_no)
    df['station_id'] = station_no
    df.loc[:,'resultUnit'] = df['resultUnit'].str.lower()
    df.replace({'resultUnit':'kg'},'lb',inplace=True)
    df.replace({'resultUnit':'ug/l'},'mg/l',inplace=True)
    df.replace({'resultUnit':'deg c'},'degF',inplace=True)
    df.replace({'resultUnit':'deg c'},'degF',inplace=True)
    
    return df.drop_duplicates(subset = 'station_id')
  

# def _info(station_nos):
#     station_info = info(station_nos)
#     if station_info.empty:
#         return Station(station_nos,
#                        'equis',
#                        station_type = 'River')
#     else:             
#         return Station(station_info.iloc[0]['stationId'],
#                        'equis',
#                        station_name = station_info.iloc[0]['stationName'],
#                        station_type = 'River')
            
    

def transform(df):
    df = df.loc[df['parameter'].isin(CONSTITUENT_MAP.keys()),:]
    df['datetime'] = pd.to_datetime(list(df.loc[:,'sampleDate'] +' ' + df.loc[:,'sampleTime']))
    df = df.loc[(df['datetime'] > '1996') & (df['result'] != '(null)')]
    
    if df.empty:
        return df
    
    df['result'] = pd.to_numeric(df['result'], errors='coerce')
    df.rename(columns = {'result': 'value',
                           'parameter': 'variable',
                           'stationName': 'station_name',
                           'stationID': 'station_id',
                           'resultUnit':'unit'},inplace=True)
    
    df['constituent'] = df['variable'].map(CONSTITUENT_MAP)
    df['station_origin'] = 'swd'
    df['quality_id'] = pd.NA
    station_name = df.iloc[0]['station_name']
    df = df.loc[:,['datetime','value','variable','unit','station_id','station_name','constituent','station_origin']]
    
    df = df.astype({'value':float,
               'unit':str,
               'station_id':str,
               'station_name':str,
               'constituent':str})
    
    # convert ug to mg/l
    df.loc[:,'unit'] = df['unit'].str.lower()
    df.loc[df['unit'] == 'ug/l','value'] = df.loc[df['unit'] == 'ug/l','value']*.001
    df.loc[df['unit'] == 'kg','value'] = df.loc[df['unit'] == 'kg','value']*2.20462
    df.loc[df['unit'] == 'deg c','value'] = df.loc[df['unit'] == 'deg c','value']*9/5 + 32 # Convert celsius to faren

    df.replace({'unit':'kg'},'lb',inplace=True)
    df.replace({'unit':'ug/l'},'mg/l',inplace=True)
    df.replace({'unit':'deg c'},'degF',inplace=True)

    # df['unit'].replace('kg','lb',inplace=True)
    # df['unit'].replace('ug/l','mg/l',inplace=True)
    # df['unit'].replace('deg c','degF',inplace=True)
    df['data_type'] = 'discrete'
    df['data_format'] = 'instantaneous'
    df.set_index('datetime',drop=True,inplace=True)
    df.index = df.index.tz_localize('UTC-06:00')
    
    df.index = df.index.round('h').round('h')
    df = df.reset_index()
    df = df.groupby(['datetime','variable','unit','station_id','station_name','constituent','data_format','data_type','station_origin']).mean()
    df = df.reset_index()
    df = df.set_index('datetime')
    df['quality_id'] = pd.NA
    df['station_name'] = station_name
    return df

def load(df,file_path):
    '''
    date, time, value, variable, unit, station_id, station_name, constituent, station_origin, data_format, data_type, quality_code, 
    
    
    
    '''


    df.to_csv(file_path)      
    
    


# base_url = 'https://webapp.pca.state.mn.us/surface-water/search?'


# https://services.pca.state.mn.us/api/v1/surfacewater/monitoring-stations/results?


# dataType
# geographicType
# specificGeoAreaCode
# wuType
# stationType
# stationId


        
# CONSTITUENT_MAP = {'TSS': ['Total suspended solids'],
#                 'TKN': ['Kjeldahl nitrogen as N','Nitrogen, Total Kjeldahl (TKN) as N'],
#                 'N'  :  ['Nitrate + Nitrite Nitrogen, Total as N','Nitrate/Nitrite as N (N+N) as N'],
#                 'TP' :  ['Phosphorus, Total as P as P'],
#                 'BOD': ['Carbonaceous biochemical oxygen demand, standard conditions',
#                                 'Chemical oxygen demand'],
#                 'CHLA': ['Chlorophyll a, corrected for pheophytin',
#                               'Chlorophyll-A',
#                               'Chlorophyll-a, Pheophytin Corrected'],
#                 'Q': ['Flow']}





from datetime import datetime, timezone, timedelta
import pandas as pd
from typing import Union
import oracledb
import duckdb

CONNECTION = None

CAS_RN_MAP = {'479-61-8':'CHLA',
            'CHLA-CORR':'CHLA',
            'BOD':'BOD',
            'NO2NO3':'N', #TODO change to 'NO2NO3'
            '14797-55-8': 'NO3',
            '14797-65-0':'NO2',
            '14265-44-2': 'OP',
            'N-KJEL' : 'TKN',
            'PHOSPHATE-P': 'TP',
            '7723-14-0' : 'TP',
            'SOLIDS-TSS': 'TSS',
            'TEMP-W' : 'WT',
            '7664-41-7' : 'NH3'}

def connect(user: str, password: str, host: str = "DELTAT", port: int = 1521, sid: str = "DELTAT"):
    '''Create and return an Oracle database connection.'''
    
    global CONNECTION
    CONNECTION = oracledb.connect(user=user, 
                                 password=password, 
                                 host=host, 
                                 port=port, 
                                 sid=sid) 
    return CONNECTION

def close_connection():
    '''Close the global Oracle database connection if it exists.'''
    global CONNECTION
    if CONNECTION:
        CONNECTION.close()
        CONNECTION = None



def test_connection():
    raise NotImplementedError("This function is a placeholder for testing Oracle DB connection.")
    try:
        # or for SID:
        # connection = oracledb.connect(user="your_username", 
        #                             password="your_password", 
        #                             host="your_host", 
        #                             port=1521, 
        #                             sid="your_sid")

        print("Successfully connected to Oracle Database")

        # Perform database operations here
        # ...
        if connection:
            connection.close()
            print("Connection closed")
    except oracledb.Error as e:
        print(f"Error connecting to Oracle Database: {e}")



def make_placeholders(items):
    '''Create SQL placeholders and bind values for a list of items'''
    # Create placeholders like :id0, :id1, :id2
    placeholders = ', '.join(f':id{i}' for i in range(len(items)))
    # Create dictionary of bind values
    binds = {f'id{i}': val for i, val in enumerate(items)}
    return placeholders, binds

def to_dataframe(odb_cursor):
    '''Convert Oracle cursor results to a pandas DataFrame'''
    column_names = [description[0] for description in odb_cursor.description]
    rows = odb_cursor.fetchall()
    df = pd.DataFrame(rows,columns = column_names)
    return df

#%% Query for station locations with HSPF related constituents
 
def download(station_ids):
    '''Download data for given station IDs from Oracle database.
    This grabs data from the Data access Layer (DAL) equis result view for
    river/stream locations and HSPF related constituents only.'''
    placeholders, binds = make_placeholders(station_ids)
    query = f"""
SELECT
    mpca_dal.eq_fac_station.latitude,
    mpca_dal.eq_fac_station.longitude,
    mpca_dal.eq_fac_station.wid_list,
    mpca_dal.eq_sample.sample_method,
    mpca_dal.eq_sample.sample_remark,
    mpca_dal.mv_eq_result.*
    FROM
    	mpca_dal.mv_eq_result
		LEFT JOIN mpca_dal.eq_fac_station 
		   ON mpca_dal.mv_eq_result.sys_loc_code = mpca_dal.eq_fac_station.sys_loc_code
		   AND mpca_dal.mv_eq_result.facility_id = mpca_dal.eq_fac_station.facility_id        
		LEFT JOIN mpca_dal.eq_sample ON mpca_dal.mv_eq_result.sample_id = mpca_dal.eq_sample.sample_id
    WHERE
        mpca_dal.mv_eq_result.cas_rn IN ('479-61-8',
                            'CHLA-CORR',
                            'BOD',
                            'NO2NO3',
                            '14797-55-8',
                            '14797-65-0',
                            '14265-44-2',
                            'N-KJEL',
                            'PHOSPHATE-P',
                            '7723-14-0',
                            'SOLIDS-TSS',
                            'TEMP-W',
                            '7664-41-7',
                            'FLOW')
        AND mpca_dal.eq_fac_station.loc_type = 'River/Stream'
        AND mpca_dal.mv_eq_result.approval_code = 'Final'
        AND mpca_dal.mv_eq_result.reportable_result = 'Y'
        AND mpca_dal.mv_eq_result.facility_id IN ( 1, 33836701 )
        AND mpca_dal.eq_sample.sample_method IN ('G-EVT', 'G', 'FIELDMSROBS', 'LKSURF1M', 'LKSURF2M', 'LKSURFOTH')
        AND mpca_dal.mv_eq_result.sys_loc_code IN ({placeholders})
    """
    with CONNECTION.cursor() as cursor:
        cursor.execute(query,binds)
        return to_dataframe(cursor)
    


def as_utc_offset(naive_dt: Union[datetime, str], tz_label: str, target_offset: timezone) -> datetime:
    """
    Interpret naive_dt (datetime or ISO string) using tz_label ('CST'|'CDT' or other).
    - If tz_label == 'CST' -> interpret as UTC-6
    - If tz_label == 'CDT' -> interpret as UTC-5
    - Else -> attempt to interpret using America/Chicago (ZoneInfo) which applies DST rules.

    WARNING : This function uses a replace mthod that assumes the input datetime is naive (no tzinfo).
    If the input datetime already has tzinfo, this will lead to incorrect results.

    Returns a timezone-aware datetime converted to a fixed UTC-6 timezone (tzinfo=UTC_MINUS_6).
    This preserves the instant and expresses it in UTC-6. 


    """
    if isinstance(naive_dt, str):
        naive = pd.to_datetime(naive_dt).to_pydatetime()
    elif isinstance(naive_dt, datetime):
        naive = naive_dt
    else:
        raise TypeError("naive_dt must be datetime or str")

    label = (tz_label or "").strip().upper()

    if label == "CST":
        src_tz = timezone(timedelta(hours=-6))
    elif label == "CDT":
        src_tz = timezone(timedelta(hours=-5))
    elif label == 'UTC':
        src_tz = timezone.utc
    else:
        raise ValueError(f"Unexpected timezone label: {tz_label}")
    # attach the source tz (interpret naive as local time in src_tz)
    aware_src = naive.replace(tzinfo=src_tz)

    # convert the instant to fixed UTC-6
    return aware_src.astimezone(target_offset).tz_localize(None)


def normalize_columns(df):
    '''Select relevant columns from Equis data.'''
    return df[['SYS_LOC_CODE',
               'constituent',
               'CAS_RN',
               'datetime',
               'RESULT_NUMERIC',
               'RESULT_UNIT',
               ]].rename(columns={
                   'SYS_LOC_CODE':'station_id',
                   'RESULT_NUMERIC':'value',
                   'RESULT_UNIT':'unit',
                   'CAS_RN':'cas_rn'
               })



def normalize_timezone(df):
    '''Normalize datetime to UTC in Equis data.'''
    target_offset = timezone(timedelta(hours=-6))
    def _conv(row):
        try:
            return as_utc_offset(row['SAMPLE_DATE_TIME'], row['SAMPLE_DATE_TIMEZONE'],target_offset)
        except Exception:
            return pd.NaT

    df.loc[:,'datetime'] = df.apply(_conv, axis=1)
    return df

def convert_units(df):
    '''Convert units in Equis data to standard units.'''
    # Convert ug/L to mg/L
    df['unit'] = df['unit'].str.lower()

    mask_ugL = df['unit'] == 'ug/l'
    df.loc[mask_ugL, 'value'] = df.loc[mask_ugL, 'value'] / 1000
    df.loc[mask_ugL, 'unit'] = 'mg/l'

    # Convert mg/g to mg/L (assuming density of 1 g/mL)
    mask_mgg = df['unit'] == 'mg/g'
    df.loc[mask_mgg, 'value'] = df.loc[mask_mgg, 'value'] * 1000
    df.loc[mask_mgg, 'unit'] = 'mg/l'

    # Convert deg C to degF
    mask_degC = df['unit'].isin(['deg c', 'degc'])
    df.loc[mask_degC, 'value'] = (df.loc[mask_degC, 'value'] * 9/5) + 32
    df.loc[mask_degC, 'unit'] = 'degf'

    return df

def map_constituents(df):
    '''Map CAS_RN to standard constituent names in Equis data.'''
    df['constituent'] = df['CAS_RN'].map(CAS_RN_MAP)
    return df


def average_results(df):
    ''' Average samples by hour, station, and constituent'''
    df['datetime'] = df['datetime'].dt.round('h')
    df['station_origin'] = 'equis'
    return df.groupby(['station_id', 'datetime', 'constituent', 'unit','station_origin']).agg(
        value=('value', 'mean')
    ).reset_index()

def replace_nondetects(df):
    '''Replace non-detect results with 0 in Equis data.'''
    df.loc[df['value'].isna(), 'value'] = 0
    return df

def normalize(df):
    '''Normalize Equis data: select relevant columns.'''
    df = map_constituents(df)
    df = normalize_timezone(df)
    df = normalize_columns(df)
    df = convert_units(df)
    return df

def transform(df):
    '''Transform Equis data: handle non-detects, convert units, map constituents.'''
    
    df = normalize(df)
    df = replace_nondetects(df)
    if not df.empty:
        df = average_results(df)
    return df



#%% Transformations using duckdb instead of pandas
# def transform_staging_to_hourly_cte(con: duckdb.DuckDBPyConnection,
#                                     source_table: str,
#                                     analytics_table: str):
#     """
#     Single-statement transformation using chained CTEs.
#     - Good when you want the whole logical pipeline in one place and avoid intermediate objects.
#     - Produces analytics.<analytics_table> as the final materialized table.
#     """

#     mapping_cases = " ".join([f"WHEN '{k}' THEN '{v}'" for k, v in CAS_RN_MAP.items()])
#     target_offset_hours = -6
#     # Example assumes source_table has: station_id, datetime, value (numeric), constituent, unit, station_origin
#     sql = f"""
#     CREATE OR REPLACE TABLE {analytics_table} AS
#     WITH
#         -- Step 1: normalize column names
#         normalized AS (
#             SELECT *,
#                 SYS_LOC_CODE AS station_id,
#                 SAMPLE_DATE_TIME AS datetime,
#                 SAMPLE_DATE_TIMEZONE AS datetime_timezone,
#                 RESULT_NUMERIC AS value,
#                 RESULT_UNIT AS unit
#             FROM {source_table}),

#         -- map constituents
#         constituents AS (
#         SELECT
#             *,
#             CASE CAS_RN
#                 {mapping_cases}
#                 ELSE NULL
#             END AS constituent
#         FROM normalized),

#         -- Step 2: convert units
#         conversions AS (
#         SELECT *,
#             CASE 
#                 WHEN LOWER(unit) = 'ug/l' THEN value / 1000
#                 WHEN LOWER(unit) = 'mg/g' THEN value * 1000
#                 WHEN LOWER(unit) IN ('deg c', 'degc') THEN (value * 9/5) + 32
#                 ELSE value
#             END AS value,
#             CASE 
#                 WHEN LOWER(unit) = 'ug/l' THEN 'mg/L'
#                 WHEN LOWER(unit) = 'mg/g' THEN 'mg/L'
#                 WHEN LOWER(unit) IN ('deg c', 'degc') THEN 'degF'
#                 ELSE unit
#             END AS unit
#         FROM constituents),

#         -- normalize timezone
#         timezones AS (
#             SELECT *,
#                 CASE
#                     WHEN datetime_timezone = 'CST' THEN 
#                         (datetime AT TIME ZONE INTERVAL '-6 hours') AT TIME ZONE INTERVAL '{target_offset_hours} hours'
#                     WHEN datetime_timezone = 'CDT' THEN 
#                         (datetime AT TIME ZONE INTERVAL '-5 hours') AT TIME ZONE INTERVAL '{target_offset_hours} hours'
#                     ELSE 
#                         datetime AT TIME ZONE INTERVAL '{target_offset_hours} hours'
#                 END AS datetime
#             FROM conversions),

        
#         hourly AS (
#                 SELECT 
#                     station_id,
#                     DATE_TRUNC('hour', datetime + INTERVAL '30 minute') AS datetime,
#                     constituent,
#                     unit,
#                     'equis' AS station_origin,
#                     AVG(value) AS value
#                 FROM timezone
#                 GROUP BY station_id, datetime, constituent, unit
#                 )

#     SELECT * FROM hourly
#     """
#     con.execute(sql)
#     return 0





# #%%

# def normalize_columns(con: duckdb.DuckDBPyConnection, table_name: str):
#     '''
#     Select relevant columns from Equis data using DuckDB.
#     '''
#     con.execute(f"""
#         CREATE TEMP VIEW v_normalized AS
#         SELECT *,
#             SYS_LOC_CODE AS station_id,
#             SAMPLE_DATE_TIME AS datetime,
#             SAMPLE_DATE_TIMEZONE AS datetime_timezone,
#             RESULT_NUMERIC AS value,
#             RESULT_UNIT AS unit
#         FROM {table_name} e
#     """)


# def map_constituents_duckdb(con: duckdb.DuckDBPyConnection, table_name: str):
#     '''
#     Map CAS_RN to standard constituent names in Equis data using DuckDB.
#     '''
    
#     mapping_cases = " ".join([f"WHEN '{k}' THEN '{v}'" for k, v in CAS_RN_MAP.items()])
#     con.execute(f"""
#         CREATE TEMP VIEW v_constituents AS
#         SELECT
#             *,
#             CASE CAS_RN
#                 {mapping_cases}
#                 ELSE NULL
#             END AS constituent
#         FROM v_normalized
#     """)

# def convert_units_duckdb(con: duckdb.DuckDBPyConnection, table_name: str):
#     '''
#     Convert units in Equis data to standard units using DuckDB.
#     '''

#     mapping_cases = " ".join([f"WHEN '{k}' THEN '{v}'" for k, v in CAS_RN_MAP.items()])
#     target_offset = timedelta(hours=-6)


#     con.execute(f"""
#         CREATE TEMP VIEW v_conversions AS
#         SELECT
#             *,


#             CASE 
#                 WHEN LOWER(unit) = 'ug/l' THEN value / 1000
#                 WHEN LOWER(unit) = 'mg/g' THEN value * 1000
#                 WHEN LOWER(unit) IN ('deg c', 'degc') THEN (value * 9/5) + 32
#                 ELSE value
#             END AS value,
#             CASE 
#                 WHEN LOWER(unit) = 'ug/l' THEN 'mg/L'
#                 WHEN LOWER(unit) = 'mg/g' THEN 'mg/L'
#                 WHEN LOWER(unit) IN ('deg c', 'degc') THEN 'degF'
#                 ELSE unit
#             END AS unit
#         FROM v_constituents""")


# def normalize_timezone(con: duckdb.DuckDBPyConnection, source_table: str, target_offset_hours: int = -6):

#     con.execute(f"""
#         CREATE TEMP VIEW v_timezone AS
#             SELECT *,
#                 CASE
#                     WHEN SAMPLE_DATE_TIMEZONE = 'CST' THEN 
#                         (SAMPLE_DATE_TIME AT TIME ZONE INTERVAL '-6 hours') AT TIME ZONE INTERVAL '{target_offset_hours} hours'
#                     WHEN SAMPLE_DATE_TIMEZONE = 'CDT' THEN 
#                         (SAMPLE_DATE_TIME AT TIME ZONE INTERVAL '-5 hours') AT TIME ZONE INTERVAL '{target_offset_hours} hours'
#                     ELSE 
#                         SAMPLE_DATE_TIME AT TIME ZONE INTERVAL '{target_offset_hours} hours'
#                 END AS datetime
#             FROM {source_table}""")


# def average_results(con: duckdb.DuckDBPyConnection, table_name: str):
#     '''
#     Average samples by hour, station, and constituent using DuckDB.
#     '''
#     con.execute(f"""
#         CREATE TABLE analytics.equis v_averaged AS
#         SELECT 
#             station_id,
#             DATE_TRUNC('hour', datetime) AS datetime,
#             constituent,
#             unit,
#             'equis' AS station_origin,
#             AVG(value) AS value
#         FROM v_timezone
#         GROUP BY station_id, DATE_TRUNC('hour', datetime), constituent, unit
#     """ )

def fetch_station_locations():
    '''Fetch station location data for stations with HSPF related constituents.'''
    query ="""SELECT DISTINCT
    m.SYS_LOC_CODE,
    stn.LONGITUDE,
    stn.LATITUDE,
    stn.LOC_MAJOR_BASIN,
    stn.NON_PUBLIC_LOCATION_FLAG
    FROM MPCA_DAL.MV_EQ_RESULT m
    LEFT JOIN MPCA_DAL.EQ_FAC_STATION_NP stn
    ON m.SYS_LOC_CODE = stn.SYS_LOC_CODE
    WHERE m.LOC_TYPE = 'River/Stream'
    AND m.CAS_RN IN ('479-61-8',
                        'CHLA-CORR',
                        'BOD',
                        'NO2NO3',
                        '14797-55-8',
                        '14797-65-0',
                        '14265-44-2',
                        'N-KJEL',
                        'PHOSPHATE-P',
                        '7723-14-0',
                        'SOLIDS-TSS',
                        'TEMP-W',
                        '7664-41-7')
        """
    with CONNECTION.cursor() as cursor:
        cursor.execute(query)
        df = to_dataframe(cursor)
        
        # dups = set(df.loc[df['SYS_LOC_CODE'].isin(df.loc[df['SYS_LOC_CODE'].duplicated()]['SYS_LOC_CODE']),'SYS_LOC_CODE'].to_list())
        # for dup in dups:
        #     #percent difference between lat/long values
        #     sub = df.loc[df['SYS_LOC_CODE'] == dup]
        #     lat_diff = abs(sub['LATITUDE'].max() - sub['LATITUDE'].min()) / ((sub['LATITUDE'].max() + sub['LATITUDE'].min()) / 2) * 100
        #     long_diff = abs(sub['LONGITUDE'].max() - sub['LONGITUDE'].min()) / ((sub['LONGITUDE'].max() + sub['LONGITUDE'].min()) / 2) * 100
        #     print(f'Duplicate station {dup} has {lat_diff:.6f}% latitude difference')
        #     print(f'Duplicate station {dup} has {long_diff:.6f}% longitude difference')

        
        # geometry = gpd.points_from_xy(df['LONGITUDE'], df['LATITUDE'])
        # gdf = gpd.GeoDataFrame(df, geometry=geometry, crs="EPSG:4326")
        # filename = 'EQ_STATION_' + str(date.today()) + '.gpkg'
        # gdf.to_file(save_path.joinpath(filename), driver = 'GPKG')
        # gdf.rename(columns={'SYS_LOC_CODE':'station_id'}, inplace=True)



import pandas as pd
from pathlib import Path

WISKI_EQUIS_XREF = pd.read_csv(Path(__file__).parent/'data/WISKI_EQUIS_XREF.csv')
#WISKI_EQUIS_XREF = pd.read_csv('C:/Users/mfratki/Documents/GitHub/hspf_tools/WISKI_EQUIS_XREF.csv')


def are_lists_identical(nested_list):
    # Sort each sublist
    sorted_sublists = [sorted(sublist) for sublist in nested_list]
    # Compare all sublists to the first one
    return all(sublist == sorted_sublists[0] for sublist in sorted_sublists)                                                                                               

def get_wiski_stations():
    return list(WISKI_EQUIS_XREF['WISKI_STATION_NO'].unique())

def get_equis_stations():
    return list(WISKI_EQUIS_XREF['EQUIS_STATION_ID'].unique())

def wiski_equis_alias(wiski_station_id):
    equis_ids =  list(set(WISKI_EQUIS_XREF.loc[WISKI_EQUIS_XREF['WISKI_STATION_NO'] == wiski_station_id,'WISKI_EQUIS_ID'].to_list()))
    equis_ids = [equis_id for equis_id in equis_ids if not pd.isna(equis_id)]
    if len(equis_ids) == 0:
        return []
    elif len(equis_ids) > 1:
        print(f'Too Many Equis Stations for {wiski_station_id}')
        raise 
    else:
        return equis_ids[0]

def wiski_equis_associations(wiski_station_id):
    equis_ids =  list(WISKI_EQUIS_XREF.loc[WISKI_EQUIS_XREF['WISKI_STATION_NO'] == wiski_station_id,'EQUIS_STATION_ID'].unique())
    equis_ids =  [equis_id for equis_id in equis_ids if not pd.isna(equis_id)]
    if len(equis_ids) == 0:
        return []
    else:
        return equis_ids
    
def equis_wiski_associations(equis_station_id):
    wiski_ids = list(WISKI_EQUIS_XREF.loc[WISKI_EQUIS_XREF['EQUIS_STATION_ID'] == equis_station_id,'WISKI_STATION_NO'].unique())
    wiski_ids = [wiski_id for wiski_id in wiski_ids if not pd.isna(wiski_id)]
    if len(wiski_ids) == 0:
        return []
    else:
        return wiski_ids
    
def equis_wiski_alias(equis_station_id):
    wiski_ids =  list(set(WISKI_EQUIS_XREF.loc[WISKI_EQUIS_XREF['WISKI_EQUIS_ID'] == equis_station_id,'WISKI_STATION_NO'].to_list()))
    wiski_ids = [wiski_id for wiski_id in wiski_ids if not pd.isna(wiski_id)]
    if len(wiski_ids) == 0:
        return []
    elif len(wiski_ids) > 1:
        print(f'Too Many WISKI Stations for {equis_station_id}')
        raise ValueError(f'Too Many WISKI Stations for {equis_station_id}')
    else:
        return wiski_ids[0]

def _equis_wiski_associations(equis_station_ids):
    wiski_stations = [equis_wiski_associations(equis_station_id) for equis_station_id in equis_station_ids]
    if are_lists_identical(wiski_stations):
        return wiski_stations[0]
    else:
        return []
        
def _stations_by_wid(wid_no,station_origin):
    if station_origin in ['wiski','wplmn']:
        station_col = 'WISKI_STATION_NO'
    elif station_origin in ['equis','swd']:
        station_col = 'EQUIS_STATION_ID'
    else:
        raise
        
    return list(WISKI_EQUIS_XREF.loc[WISKI_EQUIS_XREF['WID'] == wid_no,station_col].unique())




# -*- coding: utf-8 -*-
"""
Created on Fri Jun  3 10:01:14 2022

@author: mfratki
"""

import pandas as pd
#from abc import abstractmethod
from pathlib import Path
from mpcaHydro import etlSWD
from mpcaHydro import equis, wiski, warehouse
from mpcaHydro import xref
from mpcaHydro import outlets
from mpcaHydro.reports import reportManager
import duckdb

AGG_DEFAULTS = {'cfs':'mean',
                'mg/l':'mean',
                'degf': 'mean',
                'lb':'sum'}

UNIT_DEFAULTS = {'Q': 'cfs',
                 'QB': 'cfs',
                 'TSS': 'mg/l',
                 'TP' : 'mg/l',
                 'OP' : 'mg/l',
                 'TKN': 'mg/l',
                 'N'  : 'mg/l',
                 'WT' : 'degf',
                 'WL' : 'ft'}


def validate_constituent(constituent):
    assert constituent in ['Q','TSS','TP','OP','TKN','N','WT','DO','WL','CHLA']

def validate_unit(unit):
    assert(unit in ['mg/l','lb','cfs','degF'])


def build_warehouse(folderpath):
    folderpath = Path(folderpath)
    db_path = folderpath.joinpath('observations.duckdb').as_posix()
    warehouse.init_db(db_path)

def constituent_summary(db_path):
    with duckdb.connect(db_path) as con:
        query = '''
        SELECT
          station_id,
          station_origin,
          constituent,
          COUNT(*) AS sample_count,
          year(MIN(datetime)) AS start_date,
          year(MAX(datetime)) AS end_date
        FROM
          observations
        GROUP BY
          constituent, station_id,station_origin
        ORDER BY
          sample_count;'''
          
        res = con.execute(query)
        return res.fetch_df()


class dataManager():

    def __init__(self,folderpath, oracle_user = None, oracle_password =None):
        
        self.data = {}
        self.folderpath = Path(folderpath)
        self.db_path = self.folderpath.joinpath('observations.duckdb')
        
        self.oracle_user = oracle_user
        self.oracle_password = oracle_password
        warehouse.init_db(self.db_path,reset = False)
        self.xref = xref
        self.outlets = outlets
        self.reports = reportManager(self.db_path)

    
    def connect_to_oracle(self):
        assert (self.credentials_exist(), 'Oracle credentials not found. Set ORACLE_USER and ORACLE_PASSWORD environment variables or use swd as station_origin')
        equis.connect(user = self.oracle_user, password = self.oracle_password)
    
    def credentials_exist(self):
        if (self.oracle_user is not None) & (self.oracle_password is not None):
            return True
        else:
            return False
        
    def _build_warehouse(self):
        build_warehouse(self.folderpath)

    def download_station_data(self,station_id,station_origin,overwrite=True,to_csv = False,filter_qc_codes = True, start_year = 1996, end_year = 2030,baseflow_method = 'Boughton'):
        '''
        Method to download data for a specific station and load it into the warehouse.
        
        :param self: Description
        :param station_id: Station identifier
        :param station_origin: source of station data: wiski, equis, or swd
        :param overwrite: Whether to overwrite existing data
        :param to_csv: Whether to export data to CSV
        :param filter_qc_codes: Whether to filter quality control codes
        :param start_year: Start year for data download
        :param end_year: End year for data download
        :param baseflow_method: Method for baseflow calculation
        '''
        with duckdb.connect(self.db_path,read_only=False) as con:
            if overwrite:
                warehouse.drop_station_id(con,station_id,station_origin)
                warehouse.update_views(con)

            if station_origin == 'wiski':
                df = wiski.download([station_id],start_year = start_year, end_year = end_year)
                warehouse.load_df_to_staging(con,df, 'wiski_raw', replace = overwrite)
                warehouse.load_df_to_analytics(con,wiski.transform(df,filter_qc_codes = filter_qc_codes,baseflow_method = baseflow_method),'wiski') # method includes normalization
                
            elif station_origin == 'equis':
                assert (self.credentials_exist(), 'Oracle credentials not found. Set ORACLE_USER and ORACLE_PASSWORD environment variables or use swd as station_origin')
                df = equis.download([station_id])
                warehouse.load_df_to_staging(con,df, 'equis_raw',replace = overwrite)
                warehouse.load_df_to_analytics(con,equis.transform(df),'equis')

            elif station_origin == 'swd':
                df = etlSWD.download(station_id)
                warehouse.load_df_to_staging(con,df, 'swd_raw', replace = overwrite)
                warehouse.load_df_to_analytics(con,etlSWD.transform(df),'swd')
            else:
                raise ValueError('station_origin must be wiski, equis, or swd')    
    
        with duckdb.connect(self.db_path,read_only=False) as con:
            warehouse.update_views(con)

        if to_csv:
            self.to_csv(station_id)
            
        return df
    
    def get_outlets(self):
        with duckdb.connect(self.db_path,read_only=True) as con:
            query = '''
            SELECT *
            FROM outlets.station_reach_pairs
            ORDER BY outlet_id'''
            df = con.execute(query).fetch_df()
        return df
    def get_station_ids(self,station_origin = None):
        with duckdb.connect(self.db_path,read_only=True) as con:
            if station_origin is None:
                query = '''
                SELECT DISTINCT station_id, station_origin
                FROM analytics.observations'''
                df = con.execute(query).fetch_df()
            else:
                query = '''
                SELECT DISTINCT station_id
                FROM analytics.observations
                WHERE station_origin = ?'''
                df = con.execute(query,[station_origin]).fetch_df()
        
        return df['station_id'].to_list()
    

    def get_station_data(self,station_ids,constituent,agg_period = None):
        

        with duckdb.connect(self.db_path,read_only=True) as con:
            query = '''
            SELECT *
            FROM analytics.observations
            WHERE station_id IN ? AND constituent = ?'''
            df = con.execute(query,[station_ids,constituent]).fetch_df()
        
        unit = UNIT_DEFAULTS[constituent]
        agg_func = AGG_DEFAULTS[unit]

        df.set_index('datetime',inplace=True)
        df.attrs['unit'] = unit
        df.attrs['constituent'] = constituent
        if agg_period is not None:
            df = df[['value']].resample(agg_period).agg(agg_func)
            df.attrs['agg_period'] = agg_period

        df.rename(columns={'value': 'observed'}, inplace=True) 
        return df
    
    def get_outlet_data(self,outlet_id,constituent,agg_period = 'D'):
        with duckdb.connect(self.db_path,read_only=True) as con:
            query = '''
            SELECT *
            FROM analytics.outlet_observations_with_flow
            WHERE outlet_id = ? AND constituent = ?'''
            df = con.execute(query,[outlet_id,constituent]).fetch_df()    

        unit = UNIT_DEFAULTS[constituent]
        agg_func = AGG_DEFAULTS[unit]

        df.set_index('datetime',inplace=True)
        df.attrs['unit'] = unit
        df.attrs['constituent'] = constituent
        if agg_period is not None:
            df = df[['value','flow_value','baseflow_value']].resample(agg_period).agg(agg_func)
            df.attrs['agg_period'] = agg_period

        df.rename(columns={'value': 'observed',
                           'flow_value': 'observed_flow',
                           'baseflow_value': 'observed_baseflow'}, inplace=True) 
        return df

    

    def to_csv(self,station_id,folderpath = None):
        if folderpath is None:
            folderpath = self.folderpath
        else:
            folderpath = Path(folderpath)
        df = self._load(station_id)
        if len(df) > 0:
            df.to_csv(folderpath.joinpath(station_id + '.csv'))
        else:
            print(f'No {station_id} calibration data available at Station {station_id}')
        
        df.to_csv(folderpath.joinpath(station_id + '.csv'))


# class database():
#     def __init__(self,db_path):
#         self.dbm = MonitoringDatabase(db_path)
        
    
#     def get_timeseries(self,station_ds, constituent,agg_period):      
#         validate_constituent(constituent)
#         unit = UNIT_DEFAULTS[constituent]
#         agg_func = AGG_DEFAULTS[unit]
#         return odm.get_timeseries(station_id,constituent)

    
#     def get_samples(self,station_ds, constituent,agg_period):
#         validate_constituent(constituent)
#         unit = UNIT_DEFAULTS[constituent]
#         agg_func = AGG_DEFAULTS[unit]
#         return odm.get_sample(station_id,constituent)

#     def get_samples_and_timeseries(self,station_ds, constituent,agg_period)
        



# -*- coding: utf-8 -*-
"""
Created on Tue Oct 10 14:13:23 2023

@author: mfratki
"""

import pandas as pd
import requests
import zipfile
import io
# import geopandas as gpd


CONSITUENT_MAP = {'Water Temp. (C)': 'WT',
                 'Discharge (cfs)': 'Q',
                 'DO (mg/L)': 'DO'
    }

# def download(station_no):
#     # save_path = Path(save_path)
#     # file_path = save_path.joinpath('csg.csv')

#     station = station_no[1:]
#     df = pd.read_csv(f'https://maps2.dnr.state.mn.us/cgi-bin/csg.cgi?mode=dump_hydro_data_as_csv&site={station}&startdate=1996-1-1&enddate=2050-1-1')
#     df = pd.read_csv(f'https://apps.dnr.state.mn.us/csg/api/v1/download?callback=json&ids=66050001&vars=262')
#     df['station_id'] = station_no

#     return df

def download(station_no):
    station = station_no[1:]
    url = f'https://apps.dnr.state.mn.us/csg/api/v1/download?ids={station}&vars=262'
    response = requests.get(url)
    with zipfile.ZipFile(io.BytesIO(response.content)) as zip_ref:
        df = pd.read_csv(zip_ref.open(zip_ref.namelist()[0]))
        df['station_id'] = station_no
      
    return df
      

    # def process(df):
    #     
    #     df.set_index('Timestamp',inplace=True)
    #     value_variables = [column for column in df.columns if (column not in ['Site','Timestamp','station_no']) & ~(column.endswith('Quality'))]
        
    #     test = df[value_variables].resample(rule='1H', kind='interval').mean().dropna()
    #     df = df['Value'].resample(rule='1H', kind='interval').mean().to_frame()
        
def transform(data):
    data.rename(columns = {'tstamp': 'datetime',
                                'var_name': 'variable',
                                'station_no': 'station_id'}, inplace = True)

    data['unit'] = data['variable'].map({'Water Temp. (C)' : 'C',
                                                   'Discharge (cfs)' : 'cfs',
                                                   'DO (mg/L)' : 'mg/L'})   
    
    data['constituent'] = data['variable'].map({'Water Temp. (C)' : 'WT',
                                                          'Discharge (cfs)' : 'Q',
                                                          'DO (mg/L)' : 'DO'})
  
    data['datetime'] = pd.to_datetime(data['datetime'])
    data.set_index('datetime',drop=True,inplace=True)
    data.index = data.index.tz_localize('UTC-06:00')
    data.dropna(subset = 'value',inplace=True)
    data['source'] = 'csg'
    return data





def load(data,file_path):
    
    data.to_csv(file_path)



        



import pandas as pd
#from abc import abstractmethod
from pathlib import Path
from mpcaHydro import equis, wiski, warehouse
import duckdb





#%%
'''
This modules contains classes and functions to manage data downloads and storage into a local data warehouse.


'''

def get_db_path(warehouse_path:Path,db_name:str = 'observations')->Path:
    '''
    Constructs the full path to the database file within the warehouse directory.

    Parameters:
    warehouse_path (Path): The path to the warehouse directory.
    db_name (str): The name of the database file.

    Returns:
    Path: The full path to the database file.
    '''
    return Path(warehouse_path) / db_name

def construct_database(db_path:Path,db_name:str = 'observations')->Path:
    '''
    Constructs the full path to the database file within the warehouse directory.

    Parameters:
    warehouse_path (Path): The path to the warehouse directory.
    db_name (str): The name of the database file.

    Returns:
    Path: The full path to the database file.
    '''
    db_path = Path(db_path) / db_name
    warehouse.init_db(warehouse_path=db_path)






import duckdb
import pandas as pd
from pathlib import Path
from mpcaHydro import outlets

def init_db(db_path: str,reset: bool = False):
    """
    Initialize the DuckDB database: create schemas and tables.
    """
    db_path = Path(db_path)
    if reset and db_path.exists():
        db_path.unlink()

    with connect(db_path.as_posix()) as con:
        # Create all schemas
        create_schemas(con)

        # Create tables
        create_outlets_tables(con)
        create_mapping_tables(con)
        create_analytics_tables(con)

        # Create views
        #update_views(con)
        

def create_schemas(con: duckdb.DuckDBPyConnection):
    """
    Create staging, analytics, hspf, and reports schemas if they do not exist.
    """
    con.execute("CREATE SCHEMA IF NOT EXISTS staging")
    con.execute("CREATE SCHEMA IF NOT EXISTS analytics")
    con.execute("CREATE SCHEMA IF NOT EXISTS reports")
    con.execute("CREATE SCHEMA IF NOT EXISTS outlets")
    con.execute("CREATE SCHEMA IF NOT EXISTS mappings")

def create_analytics_tables(con: duckdb.DuckDBPyConnection):
    """
    Create necessary tables in the analytics schema.
    """
    con.execute("""
    CREATE TABLE IF NOT EXISTS analytics.equis (
        datetime TIMESTAMP,
        value DOUBLE,
        station_id TEXT,
        station_origin TEXT,
        constituent TEXT,
        unit TEXT
    );
    """)
    con.execute("""
    CREATE TABLE IF NOT EXISTS analytics.wiski (
        datetime TIMESTAMP,
        value DOUBLE,
        station_id TEXT,
        station_origin TEXT,
        constituent TEXT,
        unit TEXT
    );
    """)

def create_mapping_tables(con: duckdb.DuckDBPyConnection):
    """
    Create and populate tables in the mappings schema from Python dicts and CSVs.
    """
    # WISKI parametertype_id -> constituent
    wiski_parametertype_map = {
        '11522': 'TP', 
        '11531': 'TP', 
        '11532': 'TSS', 
        '11523': 'TSS',
        '11526': 'N', 
        '11519': 'N', 
        '11520': 'OP', 
        '11528': 'OP',
        '11530': 'TKN', 
        '11521': 'TKN', 
        '11500': 'Q', 
        '11504': 'WT',
        '11533': 'DO', 
        '11507': 'WL'
    }
    df_wiski_params = pd.DataFrame(wiski_parametertype_map.items(), columns=['parametertype_id', 'constituent'])
    con.execute("CREATE TABLE IF NOT EXISTS mappings.wiski_parametertype AS SELECT * FROM df_wiski_params")

    # EQuIS cas_rn -> constituent
    equis_casrn_map = {
        '479-61-8': 'CHLA', 
        'CHLA-CORR': 'CHLA', 
        'BOD': 'BOD', 
        'NO2NO3': 'N',
        '14797-55-8': 'NO3', 
        '14797-65-0': 'NO2', 
        '14265-44-2': 'OP',
        'N-KJEL': 'TKN', 
        'PHOSPHATE-P': 'TP', 
        '7723-14-0': 'TP',
        'SOLIDS-TSS': 'TSS', 
        'TEMP-W': 'WT', 
        '7664-41-7': 'NH3'
    }
    df_equis_cas = pd.DataFrame(equis_casrn_map.items(), columns=['cas_rn', 'constituent'])
    con.execute("CREATE TABLE IF NOT EXISTS mappings.equis_casrn AS SELECT * FROM df_equis_cas")

    # Load station cross-reference from CSV
    # Assumes this script is run from a location where this relative path is valid
    xref_csv_path = Path(__file__).parent / 'data/WISKI_EQUIS_XREF.csv'
    if xref_csv_path.exists():
        con.execute(f"CREATE TABLE IF NOT EXISTS mappings.station_xref AS SELECT * FROM read_csv_auto('{xref_csv_path.as_posix()}')")
    else:
        print(f"Warning: WISKI_EQUIS_XREF.csv not found at {xref_csv_path}")

    # Load wiski_quality_codes from CSV
    wiski_qc_csv_path = Path(__file__).parent / 'data/WISKI_QUALITY_CODES.csv'
    if wiski_qc_csv_path.exists():
        con.execute(f"CREATE TABLE IF NOT EXISTS mappings.wiski_quality_codes AS SELECT * FROM read_csv_auto('{wiski_qc_csv_path.as_posix()}')")
    else:
            print(f"Warning: WISKI_QUALITY_CODES.csv not found at {wiski_qc_csv_path}")

def create_outlets_tables(con: duckdb.DuckDBPyConnection):
    """
    Create tables in the outlets schema to define outlet-station-reach relationships.
    """
    con.execute("""-- schema.sql
            -- Simple 3-table design to manage associations between model reaches and observation stations via outlets.
            -- Compatible with DuckDB and SQLite.

            -- Table 1: outlets
            -- Represents a logical grouping that ties stations and reaches together.
            CREATE TABLE IF NOT EXISTS outlets.outlets (
            outlet_id TEXT PRIMARY KEY,
            repository_name TEXT NOT NULL,
            outlet_name TEXT,
            notes TEXT             -- optional: general notes about the outlet grouping
            );

            -- Table 2: outlet_stations
            -- One-to-many: outlet -> stations
            CREATE TABLE IF NOT EXISTS outlets.outlet_stations (
            outlet_id TEXT NOT NULL,
            station_id TEXT NOT NULL,
            station_origin TEXT NOT NULL,       -- e.g., 'wiski', 'equis'
            repository_name TEXT NOT NULL,  -- repository model the station is physically located in
            true_opnid TEXT NOT NULL,           -- The specific reach the station physically sits on (optional)
            comments TEXT,             -- Per-station comments, issues, etc.
            CONSTRAINT uq_station_origin UNIQUE (station_id, station_origin),
            FOREIGN KEY (outlet_id) REFERENCES outlets.outlets(outlet_id)
            );

            -- Table 3: outlet_reaches
            -- One-to-many: outlet -> reaches
            -- A reach can appear in multiple outlets, enabling many-to-many overall.
            CREATE TABLE IF NOT EXISTS outlets.outlet_reaches (
            outlet_id TEXT NOT NULL,
            reach_id TEXT NOT NULL,    -- model reach identifier (aka opind)
            repository_name TEXT NOT NULL,  -- optional: where the mapping comes from
            exclude INTEGER DEFAULT 0, -- flag to indicate if this reach should be excluded (1) or included (0)
            FOREIGN KEY (outlet_id) REFERENCES outlets.outlets(outlet_id)
            );

            -- Useful views:

            -- View: station_reach_pairs
            -- Derives the implicit many-to-many station <-> reach relationship via shared outlet_id
            CREATE VIEW IF NOT EXISTS outlets.station_reach_pairs AS
            SELECT
            s.outlet_id,
            s.station_id,
            s.station_origin,
            r.reach_id,
            r.exclude,
            r.repository_name,
            FROM outlets.outlet_stations s
            JOIN outlets.outlet_reaches r
            ON s.outlet_id = r.outlet_id;

          """)

def create_normalized_wiski_view(con: duckdb.DuckDBPyConnection):
    """
    Create a view in the database that contains normalized WISKI data.
    Units converted to standard units.
    columns renamed.
    constituents mapped.
    """
    con.execute("""
    -- Create a single view with all transformations
    CREATE OR REPLACE VIEW analytics.wiski_normalized AS
    SELECT 
        
        -- Convert °C to °F and keep other values unchanged
        CASE 
            WHEN LOWER(ts_unitsymbol) = '°c' THEN (value * 9.0 / 5.0) + 32
            WHEN ts_unitsymbol = 'kg' THEN value * 2.20462    -- Convert kg to lb
            ELSE value
        END AS value,

        -- Normalize units
        CASE 
            WHEN LOWER(ts_unitsymbol) = '°c' THEN 'degf'      -- Normalize °C to degF
            WHEN ts_unitsymbol = 'kg' THEN 'lb'              -- Normalize kg to lb
            WHEN ts_unitsymbol = 'ft³/s' THEN 'cfs'          -- Rename ft³/s to cfs
            ELSE ts_unitsymbol
        END AS unit,

        -- Normalize column names
        station_no AS station_id,                             -- Rename station_no to station_id
        Timestamp AS datetime,                                -- Rename Timestamp to datetime
        "Quality Code" AS quality_code,                      -- Rename Quality Code to quality_code
        "Quality Code Name" AS quality_code_name,            -- Rename Quality Code Name to quality_code_name
        parametertype_id,                                    -- Keeps parametertype_id as is
        constituent                                          -- Keeps constituent as is
    FROM staging.wiski_raw;""")


def create_filtered_wiski_view(con: duckdb.DuckDBPyConnection, data_codes: list):
    """
    Create a view in the database that filters WISKI data based on specified data codes.
    """
    query = f"""
    CREATE OR REPLACE VIEW analytics.wiski_filtered AS
    SELECT *
    FROM analytics.wiski_normalized
    WHERE quality_code IN ({placeholders});
    """

    placeholders = ', '.join(['?'] * len(data_codes))
    query = query.format(placeholders=placeholders)
    con.execute(query, data_codes)


def create_aggregated_wiski_view(con: duckdb.DuckDBPyConnection):
    """
    Create a view in the database that aggregates WISKI data by hour, station, and constituent.
    """
    con.execute("""
    CREATE OR REPLACE Table analytics.wiski_aggregated AS
    SELECT 
        station_id,
        constituent,
        time_bucket(INTERVAL '1 hour', datetime) AS hour_start,
        AVG(value) AS value,
        unit
    FROM analytics.wiski_normalized
    GROUP BY 
        station_id, 
        constituent, 
        hour_start,
        unit;
    """)

def create_staging_qc_count_view(con: duckdb.DuckDBPyConnection):
    """
    Create a view in staging schema that counts quality codes for each station and constituent.
    """
    con.execute("""
    CREATE OR REPLACE VIEW staging.wiski_qc_count AS (
        SELECT 
            w.station_no,
            w.parametertype_name,
            w."Quality Code",
            w."Quality Code Name",
            COUNT(w."Quality Code") AS count
        FROM staging.wiski_raw w 
        GROUP BY
            w."Quality Code",w."Quality Code Name",w.parametertype_name, w.station_no
                );
        """)
    # ORDER BY
    #         w.station_no,w.parametertype_name, w."Quality Code"
    #         )
    # """)

def create_combined_observations_view(con: duckdb.DuckDBPyConnection):
    """
    Create a view in analytics schema that combines observations from equis and wiski processed tables.
    """
    con.execute("""
    CREATE OR REPLACE VIEW analytics.observations AS
    SELECT datetime,value,station_id,station_origin,constituent,unit
    FROM analytics.equis
    UNION ALL
    SELECT datetime,value,station_id,station_origin,constituent,unit
    FROM analytics.wiski;
    """)


def create_outlet_observations_view(con: duckdb.DuckDBPyConnection):
    """
    Create a view in analytics schema that links observations to model reaches via outlets.
    """
    con.execute("""
    CREATE OR REPLACE VIEW analytics.outlet_observations AS 
    SELECT
        o.datetime,
        os.outlet_id,
        o.constituent,
        AVG(o.value) AS value,
        COUNT(o.value) AS count
    FROM
        analytics.observations AS o
    LEFT JOIN
        outlets.outlet_stations AS os ON o.station_id = os.station_id AND o.station_origin = os.station_origin
    GROUP BY
        os.outlet_id,
        o.constituent,
        o.datetime; -- Group by the truncated date
    """)
    # ORDER BY
    #     os.outlet_id,
    #     o.constituent,
    #     datetime);



def create_outlet_observations_with_flow_view(con: duckdb.DuckDBPyConnection):
    
    con.execute("""
    CREATE OR REPLACE VIEW analytics.outlet_observations_with_flow AS 
        WITH baseflow_data AS (
            SELECT
                outlet_id,
                datetime,
                "value" AS baseflow_value
            FROM
                analytics.outlet_observations
            WHERE
                (constituent = 'QB')),
        flow_data AS (
            SELECT
                outlet_id,
                datetime,
                "value" AS flow_value
            FROM
                analytics.outlet_observations
            WHERE
                (constituent = 'Q')),
        constituent_data AS (
            SELECT
                outlet_id,
                datetime,
                constituent,
                "value",
                count
            FROM
                analytics.outlet_observations
            WHERE
                (constituent NOT IN ('Q', 'QB')))
        SELECT
            constituent_data.outlet_id,
            constituent_data.constituent,
            constituent_data.datetime,
            constituent_data."value",
            flow_data.flow_value,
            baseflow_data.baseflow_value
        FROM
            constituent_data
        FULL JOIN flow_data ON
            (((constituent_data.outlet_id = flow_data.outlet_id)
                AND (constituent_data.datetime = flow_data.datetime)))
        LEFT JOIN baseflow_data ON
            (((constituent_data.outlet_id = baseflow_data.outlet_id)
                AND (constituent_data.datetime = baseflow_data.datetime)));""")
    # ORDER BY
    #     constituent_data.outlet_id,
    #     constituent_data.datetime;
    # 

def create_constituent_summary_report(con: duckdb.DuckDBPyConnection):
    """
    Create a constituent summary report in the reports schema that groups observations by constituent and station.
    """
    con.execute('''
            CREATE OR REPLACE VIEW reports.constituent_summary AS
            SELECT
            station_id,
            station_origin,
            constituent,
            COUNT(*) AS sample_count,
            AVG(value) AS average_value,
            MIN(value) AS min_value,
            MAX(value) AS max_value,
            year(MIN(datetime)) AS start_date,
            year(MAX(datetime)) AS end_date
            FROM
            analytics.observations
            GROUP BY
            constituent,station_id,station_origin;
            ''')
                
            # ORDER BY
            # constituent,sample_count;''')

def create_outlet_summary_report(con: duckdb.DuckDBPyConnection):
    con.execute("""
        CREATE VIEW reports.outlet_constituent_summary AS
    SELECT
        outlet_id,
        constituent,
        count_star() AS sample_count,
        avg("value") AS average_value,
        min("value") AS min_value,
        max("value") AS max_value,
        "year"(min(datetime)) AS start_date,
        "year"(max(datetime)) AS end_date
    FROM
        analytics.outlet_observations
    GROUP BY
        constituent,
        outlet_id
    """)

    
       
def drop_station_id(con: duckdb.DuckDBPyConnection, station_id: str,station_origin: str):
    """
    Drop all data for a specific station from staging and analytics schemas.
    """
    con.execute(f"DELETE FROM staging.equis_raw WHERE station_id = '{station_id}' AND station_origin = '{station_origin}'")
    con.execute(f"DELETE FROM staging.wiski_raw WHERE station_id = '{station_id}' AND station_origin = '{station_origin}'")
    con.execute(f"DELETE FROM analytics.equis WHERE station_id = '{station_id}' AND station_origin = '{station_origin}'")
    con.execute(f"DELETE FROM analytics.wiski WHERE station_id = '{station_id}' AND station_origin = '{station_origin}'")
    update_views(con)

def update_views(con: duckdb.DuckDBPyConnection):
    """
    Update all views in the database.
    """
    create_staging_qc_count_view(con)
    create_combined_observations_view(con)
    create_constituent_summary_report(con)
    create_outlet_observations_view(con)
    create_outlet_observations_with_flow_view(con)
    create_outlet_summary_report(con)
    
def connect(db_path: str, read_only: bool = False) -> duckdb.DuckDBPyConnection:
    """
    Returns a DuckDB connection to the given database path.
    Ensures the parent directory exists.
    """
    db_path = Path(db_path)
    parent = db_path.parent
    parent.mkdir(parents=True, exist_ok=True)
    return duckdb.connect(database=db_path.as_posix(), read_only=read_only)


def load_df_to_table(con: duckdb.DuckDBPyConnection, df: pd.DataFrame, table_name: str, replace: bool = True):
    """
    Persist a pandas DataFrame into a DuckDB table. This will overwrite the table
    by default (replace=True).
    """
    if replace:
        con.execute(f"DROP TABLE IF EXISTS {table_name}")
    # register pandas DF and create table
    con.register("tmp_df", df)
    con.execute(f"CREATE TABLE {table_name} AS SELECT * FROM tmp_df")
    con.unregister("tmp_df")

def load_df_to_staging(con: duckdb.DuckDBPyConnection, df: pd.DataFrame, table_name: str, replace: bool = True):
    """
    Persist a pandas DataFrame into a staging table. This will overwrite the staging
    table by default (replace=True).
    """
    if replace:
        con.execute(f"DROP TABLE IF EXISTS staging.{table_name}")
    # register pandas DF and create table
    con.register("tmp_df", df)
    con.execute(f"CREATE TABLE staging.{table_name} AS SELECT * FROM tmp_df")
    con.unregister("tmp_df")

def add_df_to_staging(con: duckdb.DuckDBPyConnection, df: pd.DataFrame, table_name: str):
    """
    Append a pandas DataFrame into a staging table. This will create the staging
    table if it does not exist.
    """
    # register pandas DF and create table if not exists
    con.register("tmp_df", df)
    con.execute(f"""
        CREATE TABLE IF NOT EXISTS staging.{table_name} AS 
        SELECT * FROM tmp_df
    """)
    con.execute(f"""
        INSERT INTO staging.{table_name} 
        SELECT * FROM tmp_df
    """)
    con.unregister("tmp_df")

def load_csv_to_staging(con: duckdb.DuckDBPyConnection, csv_path: str, table_name: str, replace: bool = True, **read_csv_kwargs):
    """
    Persist a CSV file into a staging table. This will overwrite the staging
    table by default (replace=True).
    """
    if replace:
        con.execute(f"DROP TABLE IF EXISTS staging.{table_name}")
    con.execute(f"""
        CREATE TABLE staging.{table_name} AS 
        SELECT * FROM read_csv_auto('{csv_path}', {', '.join(f"{k}={repr(v)}" for k, v in read_csv_kwargs.items())})
    """)

def load_parquet_to_staging(con: duckdb.DuckDBPyConnection, parquet_path: str, table_name: str, replace: bool = True):
    """
    Persist a Parquet file into a staging table. This will overwrite the staging
    table by default (replace=True).
    """
    if replace:
        con.execute(f"DROP TABLE IF EXISTS staging.{table_name}")
    con.execute(f"""
        CREATE TABLE staging.{table_name} AS 
        SELECT * FROM read_parquet('{parquet_path}')
    """)


def write_table_to_parquet(con: duckdb.DuckDBPyConnection, table_name: str, path: str, compression="snappy"):
    """
    Persist a DuckDB table into a Parquet file.
    """
    con.execute(f"COPY (SELECT * FROM {table_name}) TO '{path}' (FORMAT PARQUET, COMPRESSION '{compression}')")


def write_table_to_csv(con: duckdb.DuckDBPyConnection, table_name: str, path: str, header: bool = True, sep: str = ',', **kwargs):
    """
    Persist a DuckDB table into a CSV file.
    """
    con.execute(f"COPY (SELECT * FROM {table_name}) TO '{path}' (FORMAT CSV, HEADER {str(header).upper()}, DELIMITER '{sep}' {', '.join(f', {k}={repr(v)}' for k, v in kwargs.items())})")




def load_df_to_analytics(con: duckdb.DuckDBPyConnection, df: pd.DataFrame, table_name: str):
    """
    Persist a pandas DataFrame into an analytics table.
    """
    con.execute(f"DROP TABLE IF EXISTS analytics.{table_name}")
    con.register("tmp_df", df)
    con.execute(f"CREATE TABLE analytics.{table_name} AS SELECT * FROM tmp_df")
    con.unregister("tmp_df")


def migrate_staging_to_analytics(con: duckdb.DuckDBPyConnection, staging_table: str, analytics_table: str):
    """
    Migrate data from a staging table to an analytics table.
    """
    con.execute(f"DROP TABLE IF EXISTS analytics.{analytics_table}")
    con.execute(f"""
        CREATE TABLE analytics.{analytics_table} AS 
        SELECT * FROM staging.{staging_table}
    """)


def load_to_analytics(con: duckdb.DuckDBPyConnection, table_name: str):
    con.execute(f"""
                CREATE OR REPLACE TABLE analytics.{table_name} AS
                SELECT
                station_id,
                constituent,
                datetime,
                value AS observed_value,
                time_bucket(INTERVAL '1 hour', datetime) AS hour_start,
                AVG(observed_value) AS value
                FROM
                    staging.equis_processed
                GROUP BY
                    hour_start,
                    constituent,
                    station_id
                ORDER BY
                    station_id,
                    constituent
                """)
    # register pandas DF and create table
    con.register("tmp_df", df)
    con.execute(f"CREATE TABLE analytics.{table_name} AS SELECT * FROM tmp_df")
    con.unregister("tmp_df")

def dataframe_to_parquet(con: duckdb.DuckDBPyConnection,  df: pd.DataFrame, path, compression="snappy"):
    # path should be a filename like 'data/raw/equis/equis-20251118.parquet'
    con = duckdb.connect()
    con.register("tmp_df", df)
    con.execute(f"COPY (SELECT * FROM tmp_df) TO '{path}' (FORMAT PARQUET, COMPRESSION '{compression}')")
    con.unregister("tmp_df")
    con.close()


# -*- coding: utf-8 -*-
"""
Created on Mon Jul 10 16:18:03 2023

@author: mfratki
"""
import requests
from requests.exceptions import ConnectionError, Timeout, HTTPError, RequestException
import pandas as pd
import time

#TODO: Use this url to make sure web service is working https://wiskiweb01.pca.state.mn.us/
class Service():
    base_url = 'http://wiskiweb01.pca.state.mn.us/KiWIS/KiWIS?'
    base_dict = {
        'datasource': '0',
        'service': 'kisters',
        'type': 'queryServices',
        'format': 'json'}
    
    def __init__(self):
        self._url = None
        self._args = None
        self.response = None
        
    def test_connection(self):
        timeout = 5 
        try:
            # Using requests.head() to fetch headers is faster than requests.get()
            # as it doesn't download the full content
            response = requests.head('http://wiskiweb01.pca.state.mn.us', timeout=timeout)
            
            # raise_for_status() raises an HTTPError for 4xx or 5xx status codes
            response.raise_for_status() 
            
            # If no exception was raised, the website is considered "up"
            return True, f"Website is UP (Status Code: {response.status_code})"
        
        except ConnectionError as e:
            # Handles DNS failures, refused connections, etc.
            return False, f"Website is DOWN (Connection Error): {e}"
        except Timeout as e:
            # Handles cases where the server takes too long to respond
            return False, f"Website is DOWN (Timeout Error): {e}"
        except HTTPError as e:
            # Handles HTTP errors like 404 Not Found, 500 Internal Server Error, etc.
            return False, f"Website is experiencing issues (HTTP Error): {e}"
        except RequestException as e:
            # Handles any other exceptions that might occur during the request
            return False, f"An unexpected error occurred: {e}"

    def _requestTypes(self):
        url = self.url({'request': 'getrequestinfo'}) 
        return requests.get(url).json()[0]  
        
    def getRequests(self):
        return list(self._requestTypes()['Requests'].keys())
     
    def queryfields(self,request_type):
        return list(self._requestTypes()['Requests'][request_type]['QueryFields']['Content'].keys())
     
    def returnfields(self,request_type):
        return list(self._requestTypes()['Requests'][request_type]['Returnfields']['Content'].keys())
    
    def optionalfields(self,request_type):
        return list(self._requestTypes()['Requests'][request_type]['Optionalfields']['Content'].keys())
    
    def formats(self,request_type):
        return list(self._requestTypes()['Requests'][request_type]['Formats']['Content'].keys())
    
    def info(self,request_type):
        url = self.url({'request': 'getrequestinfo'})
        response = requests.get(url)
        get_requests = response.json()    
        return get_requests[0]['Requests'].keys()
        

    def url(self,args_dict):
        args_dict = self.base_dict | args_dict
        args = []
        for k,v in args_dict.items():
            if v is None:
                continue
            elif isinstance(v,list):
                v = [str(vv) for vv in v]
                v = ','.join(v)    
            args.append(f'{k}={v}')
        args = '&'.join(args)
        
        url = self.base_url + args
        self._url = url
        return url
    
    def get_json(self,args_dict):
        # Download request
        self.response = requests.get(self.url(args_dict))
        if self.response.status_code != 200:
            print('Error: ' + self.response.json()['message'])
            self.response.raise_for_status()  # raises exception when not a 2xx response

        return  self.response.json()  
    
    def df(self,args_dict):


        get_requests = self.get_json(args_dict)
        # Convert to dataframe
        if args_dict['request'] in ['getTimeseriesValues']:
            dfs = []
            for get_request in get_requests:
                df = pd.DataFrame(get_request['data'],columns = get_request['columns'].split(','))
                del get_request['data']
                del get_request['rows']
                del get_request['columns']  
                for k,v in get_request.items(): df[k] = v       
                dfs.append(df)
            df = pd.concat(dfs)
        else:
            df = pd.DataFrame(get_requests[1:], columns = get_requests[0])
        
        # print('Done!')
        return df
    
    def get(self,args):
        request_type = args['request']
        #assert(request_type in self.getRequests())
        _args = {queryfield: None for queryfield in self.queryfields(request_type)} | {optionalfield: None for optionalfield in self.optionalfields(request_type)}
        args = {**_args, **args}
        self._args = args        
        return self.df(args)
    
    def _filter(self,args):
        
        '''
        Filter for ensuring not too many values are requested and determining the proper division
        given the number of timeseries, timeseries length, and timeseries sampling interval
        '''
        'minute','hour','daily'
        
        MAX_OUTPUT = 240000 #True max output is 250,000 but giving myself a bit of a buffer
        
        
        n_timeseries = 1
        n_years = 1
        #1 timeseries for 1 year
        n_values = 60*24*365*n_timeseries*n_years
        
        if n_values < MAX_OUTPUT :
            return 0
        elif n_timeseries == 1:
            n_values/MAX_OUTPUT
                

    
'''
Potential use cases:

1. timeseries for a given ts_id
2. All timeseries for a given station
3. All timeseries for a given parameter
4. All timeseries for a given huc_id    
5. All timeseries of a given resolution 

'''

''''
Aggregate (aggregate) - Builds representative periodic values.
Attributes:	Interval (MANDATORY): HHMMSS, decadal, yearly, year, monthly, month, daily, day, hourly, hour
Aggregation Types (MANDATORY LIST): min, max, mean, average, total, counts, perc-#
Returnfields:	Count
Interpolation Type
Average
Quality Code
Mean
Total
Maximum
Minimum
Timestamp
P#
Examples:	aggregate(daily~total)
aggregate(yearly~mean~min~max)
aggregate(hourly~perc-25~perc-75)
'''

VALID_AGGREGATION_TYPES = ['min', 'max', 'mean', 'average', 'total', 'counts']
VALID_INTERVALS = ['decadal', 'yearly', 'year', 'monthly', 'month', 'daily', 'day', 'hourly', 'hour']
SERVICE = Service()

def construct_aggregation(interval, aggregation_type):
    validate_interval(interval)
    validate_aggregation_type(aggregation_type) 
    return f'aggregate({interval}~{aggregation_type})'

def validate_aggregation_type(aggregation_type):
    assert(aggregation_type in VALID_AGGREGATION_TYPES or validate_percentile(aggregation_type))
    return True
   
def validate_percentile(aggregation_type):
    assert(aggregation_type.startswith('perc-'))
    perc_value = aggregation_type.split('-')[1]
    assert(perc_value.isdigit())
    perc_value = int(perc_value)
    assert(0 < perc_value < 100)
    return True


def validate_interval(interval):
    assert(interval in VALID_INTERVALS or validate_custom_interval(interval))
    return True

def validate_custom_interval(interval:str):
    # Custom interval in HHMMSS format
    assert(len(interval) == 6)
    assert(all(char.isdigit() for char in interval))
    assert(0 <= int(interval[0:2]) < 24)  # hours
    assert(0 <= int(interval[2:4]) < 60)  # minutes
    assert(0 <= int(interval[4:6]) < 60)  # seconds
    return True


def test_connection():
    return SERVICE.test_connection()

def get(args_dict):
    return SERVICE.get(args_dict)

def get_ts(
            ts_id,
            aggregation_interval = None,
            aggregation_type = None,
            start_date = '1996-01-01',
            end_date = '2050-12-31',
            stationgroup_id = None,
            timezone = 'GMT-6',
            as_json = False):
    
    if (aggregation_interval is not None) and (aggregation_type is not None):
        transformation = construct_aggregation(aggregation_interval, aggregation_type)
        ts_id = f'{ts_id};{transformation}'

    #print('Downloading Timeseries Data')
    args = {'request':'getTimeseriesValues',
            'ts_id' : ts_id,
            'from': start_date,
            'to': end_date,
            'returnfields': ['Timestamp', 'Value', 'Quality Code','Quality Code Name'],
            'metadata': 'true',
            'md_returnfields': ['ts_unitsymbol',
                                'ts_name',
                                'ts_id',
                                'station_no',
                                'station_name',
                                'station_latitude',
                                'station_longitude',
                                'parametertype_id',
                                'parametertype_name',
                                'stationparameter_no',
                                'stationparameter_name'],
            'timezone':timezone,
            'ca_sta_returnfields': ['stn_HUC12','stn_EQuIS_ID']}
    
    if as_json:
        output = SERVICE.get_json(args)
    else: 
        output = SERVICE.get(args)
    #print('Done!')
    return output
    
def get_stations(
                    huc_id = None, 
                    parametertype_id = None,
                    stationgroup_id = None,
                    stationparameter_no = None,
                    station_no = None,
                    returnfields = []):
    
    args = {'request':'getStationList'}
    
    returnfields = list(set(['ca_sta','station_no','station_name'] + returnfields))
        
    args ={'request': 'getStationList',
            'stationparameter_no': stationparameter_no,
            'stationgroup_id': stationgroup_id,
            'parametertype_id': parametertype_id,
            'station_no': station_no,
            #'object_type': object_type,
            'returnfields': returnfields,
            #                  'parametertype_id','parametertype_name',
            #                  'station_latitude','station_longitude',
            #                  'stationparameter_no','stationparameter_name'],
            'ca_sta_returnfields': ['stn_HUC12','stn_EQuIS_ID','stn_AUID','hydrounit_title','hydrounit_no','NearestTown']
            }
    
    
    df = SERVICE.get(args)
    if huc_id is not None: df = df.loc[df['stn_HUC12'].str.startswith(huc_id)]
    return df

def get_ts_ids(
                station_nos,
                parametertype_id = None,
                stationparameter_no = None,
                stationgroup_id = None,
                ts_name = None,
                returnfields = None):
    

    if returnfields is None:
        returnfields = ['ts_id','ts_name','ca_sta','station_no',
                            'ts_unitsymbol',
                            'parametertype_id','parametertype_name',
                            'station_latitude','station_longitude',
                            'stationparameter_no','stationparameter_name',
                            'station_no','station_name',
                            'coverage','ts_density']
    

    args ={'request': 'getTimeseriesList',
            'station_no': station_nos,
            'parametertype_id': parametertype_id,
            'stationparameter_no': stationparameter_no,
            'ts_name' : ts_name,
            'returnfields': returnfields,
            'ca_sta_returnfields': ['stn_HUC12','stn_EQuIS_ID','stn_AUID']}
    
    df = SERVICE.get(args)
    return df
    


def get_wplmn(station_nos):
    
    PARAMETERS_MAP={'5004':'TP Load',
                    '5005':'TP Conc',
                    '5014':'TSS Load',
                    '5015':'TSS Conc',
                    '5024':'N Load',
                    '5025':'N Conc',
                    '5034':'OP Load',
                    '5035':'OP Conc',
                    '5044':'TKN Load',
                    '5045':'TKN Conc',
                    '262' :'Flow'}
        
    ts_ids = self.get_ts_ids(station_nos = station_nos,
                        stationgroup_id = '1319204',
                        stationparameter_no = list(PARAMETERS_MAP.keys()),
                        ts_name = ['20.Day.Mean'])
    
    if len(ts_ids) == 0:
        print('No WPLMN Sites Available')
        return pd.DataFrame() 
    
    dfs = []
    for ts_id in ts_ids['ts_id']:
        dfs.append(self.get_ts(ts_id))
        time.sleep(1)
    
    return pd.concat(dfs)


        
# nutrient
#     -N03N02
#     -OP
#     -NH3
#     -TP
#     -DO
#     -CHla
# temperature
# flow

# test = pyWISK()

# df = test.get_ts(ts_ids = 424663010)

# df = test.get_ts(station_nos = 'W25060001')

# df = test.get_wplmn(huc8_id = '07020005')

# df = test.get_ts(huc_id = '07010205',stationgroup_id = '1319204',parametertype_id = 11500)



import pandas as pd
from mpcaHydro import pywisk
import baseflow as bf
import time


#%% Define Selectors and Maps
PARAMETERTYPE_MAP ={'11522': 'TP',
                    '11531': 'TP',
                    '11532': 'TSS',
                    '11523': 'TSS',
                    '11526': 'N',
                    '11519': 'N',
                    '11520': 'OP',
                    '11528': 'OP',
                    '11530': 'TKN',
                    '11521': 'TKN',
                    '11500' : 'Q',
                    '11504': 'WT',
                    '11533': 'DO',
                    '11507':'WL'}
#STATIONPARAMETER_NOS = ['262*','450*','451*','863*','866*','5034' ,'5035','5005', '5004','5014' ,'5015','5024'  ,'5025','5044' ,'5045']
STATIONPARAMETER_NOS = ['262*','450*','451*','863*','866*']

DATA_CODES = [1,3,10,12,15,20,29,30,31,32,34,45,46,47,48,49]


TS_NAME_SELECTOR = {'Q':{'Internal':{'daily':'20.Day.Mean.Archive',
                                     'unit': '15.Rated'},
                         'External': {'daily': '20.Day.Mean',
                                      'unit': '08.Provisional.Edited'}},
                    'WT':{'Internal':{'daily':'20.Day.Mean',
                                      'unit': '09.Archive'},
                          'External': {'daily': '20.Day.Mean',
                                       'unit': '08.Provisional.Edited'}},
                    'TSS':{'Internal':{'daily':'20.Day.Mean',
                                      'unit': '09.Archive'},
                          'External': {'daily': '20.Day.Mean',
                                       'unit': '08.Provisional.Edited'}},                   
                    'N':{'Internal':{'daily':'20.Day.Mean',
                                      'unit': '09.Archive'},
                          'External': {'daily': '20.Day.Mean',
                                       'unit': '08.Provisional.Edited'}},                    
                    'TKN':{'Internal':{'daily':'20.Day.Mean',
                                      'unit': '09.Archive'},
                          'External': {'daily': '20.Day.Mean',
                                       'unit': '08.Provisional.Edited'}},                    
                    'TP':{'Internal':{'daily':'20.Day.Mean',
                                      'unit': '09.Archive'},
                          'External': {'daily': '20.Day.Mean',
                                       'unit': '08.Provisional.Edited'}},                    
                    'OP':{'Internal':{'daily':'20.Day.Mean',
                                      'unit': '09.Archive'},
                          'External': {'daily': '20.Day.Mean',
                                       'unit': '08.Provisional.Edited'}},                    
                    'DO':{'Internal':{'daily':'20.Day.Mean',
                                      'unit': '09.Archive'},
                          'External': {'daily': '20.Day.Mean',
                                       'unit': '08.Provisional.Edited'}},
                    'TRB':{'Internal':{'daily':'20.Day.Mean',
                                    'unit': '09.Archive'},
                        'External': {'daily': '20.Day.Mean',
                                    'unit': '08.Provisional.Edited'}}}



CONSTITUENT_NAME_NO = {'Q'  :['262*'],#,'263'],
                       'WT' :['450*', '451*'], # '450.42','451.42'],
                       'OP' :['863*'],
                       'DO' :['866*'],
                       'TRB':['811*'],
                       'TP' :None,
                       'TSS':None,
                       'N'  :None,
                       'TKN':None}

CONSTITUENT_NAME_NO_WPLMN = {'Q'  :['262*'],#,'263'],
                       'WT' :['450*', '451*'], # '450.42','451.42'],
                       'OP' :['863*','5034' ,'5035'],
                       'DO' :['866*'],
                       'TP' :['5005'  ,'5004'],
                       'TSS':['5014' ,'5015'],
                       'N'  :['5024'  ,'5025'],
                       'TKN':['5044' ,'5045']}

VALID_CONSTITUENTS = ['Q','WT','OP','DO','TP','TSS','N','TKN','TRB']

def test_connection():
    '''
    Test connection to WISKI database.
    '''
    return pywisk.test_connection()

def download(station_ids: list, start_year: int = 1996, end_year: int = 2030,wplmn: bool = False):
    '''
    Fetch data for given station IDs from WISKI database using the KISTERS API.
    '''
    dfs = [pd.DataFrame()]
    for station_id in station_ids:
        if not isinstance(station_id,str):
            raise ValueError(f'Station ID {station_id} is not a string')
        print('Downloading Timeseries Data')
        df = pd.concat([_download(constituent,station_id,start_year,end_year,wplmn) for constituent in VALID_CONSTITUENTS])

        if not df.empty:
            dfs.append(df)
    df = pd.concat(dfs)

    station_metadata = pywisk.get_stations(station_no = station_ids,returnfields = ['stationgroup_id'])
    if any(station_metadata['stationgroup_id'].isin(['1319204'])):
        df['wplmn_flag'] = 1
    else:
        df['wplmn_flag'] = 0
    print('Done!')
    
    return df

def _download(constituent,station_nos,start_year = 1996,end_year = 2030,wplmn = False):

    if station_nos[0] == 'E':
        ts_names = TS_NAME_SELECTOR[constituent]['External']
    else:
        ts_names =TS_NAME_SELECTOR[constituent]['Internal']
    
    if wplmn:
        constituent_nos = CONSTITUENT_NAME_NO_WPLMN[constituent]
    else:
        constituent_nos = CONSTITUENT_NAME_NO[constituent]
    
    if constituent_nos is not None:
        ts_ids = pywisk.get_ts_ids(station_nos = station_nos,
                            stationparameter_no = constituent_nos,
                            ts_name = ts_names['unit'])
        
        if ts_ids.empty:
            ts_ids = pywisk.get_ts_ids(station_nos = station_nos,
                                stationparameter_no = constituent_nos,
                                ts_name = ts_names['daily'])
            if ts_ids.empty:
                return pd.DataFrame()    
        
        df = convert_to_df(ts_ids['ts_id'],start_year,end_year)

        if df.empty:
            return pd.DataFrame()    
    else:
        df = pd.DataFrame()
    return df


def download_chunk(ts_id,start_year = 1996,end_year = 2030, interval = 4, as_json = False):
    frames = [pd.DataFrame()]

    for start in range(start_year,end_year,interval):
        end = int(start + interval-1)
        if end > end_year:
            end = end_year
        df = pywisk.get_ts(ts_id,start_date = f'{start}-01-01',end_date = f'{end}-12-31',as_json = as_json)
        if not df.empty: frames.append(df)
        df['Timestamp'] = pd.to_datetime(df['Timestamp']).dt.tz_localize(None)
        time.sleep(.1)   
    return pd.concat(frames)

def convert_to_df(ts_ids,start_year = 1996,end_year = 2030):
    dfs = []
    for ts_id in ts_ids:
        dfs.append(download_chunk(ts_id,start_year,end_year))
        time.sleep(.1)
    df =  pd.concat(dfs)
    return df


def discharge(station_nos,start_year = 1996,end_year = 2030):
    return _download('Q',station_nos,start_year,end_year)


def temperature(station_nos,start_year = 1996,end_year = 2030):
    return _download('WT',station_nos,start_year,end_year)


def orthophosphate(station_nos,start_year = 1996,end_year = 2030):
    return _download('OP',station_nos,start_year,end_year)

def dissolved_oxygen(station_nos,start_year = 1996,end_year = 2030):
    return _download('DO',station_nos,start_year,end_year)

def nitrogen(station_nos,start_year = 1996,end_year = 2030):
    return _download('N',station_nos,start_year,end_year)

def total_suspended_solids(station_nos,start_year = 1996,end_year = 2030):
    return _download('TSS',station_nos,start_year,end_year)

def total_phosphorous(station_nos,start_year = 1996,end_year = 2030):
    return _download('TP',station_nos,start_year,end_year)

def tkn(station_nos,start_year = 1996,end_year = 2030):
    return _download('TKN',station_nos,start_year,end_year)





def convert_units(df):
    '''
    Convert units to standard units
    '''
    # Convert units
    #Water temperature``
    df.loc[:,'ts_unitsymbol'] = df['ts_unitsymbol'].str.lower()
    df.replace({'ts_unitsymbol':'°c'},'degf',inplace = True)
    df.loc[df['ts_unitsymbol'] == 'degf','Value'] = df.loc[df['ts_unitsymbol'] == 'degf','Value'].apply(lambda x: (x*9/5)+32)

    # Convert kg to lb
    df.loc[df['ts_unitsymbol'] == 'kg','Value'] = df.loc[df['ts_unitsymbol'] == 'kg','Value'].apply(lambda x: (x*2.20462))
    df.replace({'ts_unitsymbol':'kg'},'lb',inplace=True)

    # rename ft3/s to cfs
    df.replace({'ts_unitsymbol':'ft³/s'},'cfs',inplace=True)
    return df


def normalize_columns(df):
    '''
    Normalize column names and units
    '''
    # Map parameter numbers to constituent names
    df['constituent'] = df['parametertype_id'].map(PARAMETERTYPE_MAP)
    
    df.rename(columns={
        'station_no':'station_id',
        'Timestamp':'datetime',
        'Value':'value',
        'ts_unitsymbol':'unit',
        'Quality Code':'quality_code',
        'Quality Code Name':'quality_code_name'}, inplace=True)
    return df
    


def filter_quality_codes(df, data_codes):
    '''
    Filter dataframe by valid quality codes
    '''
    return df.loc[df['quality_code'].isin(data_codes)]

def average_results(df):
    #df['datetime'] = pd.to_datetime(df.loc[:,'datetime'])
    df.loc[:,'datetime'] = df.loc[:,'datetime'].dt.round('h')
    return df.groupby(['station_id', 'datetime', 'constituent', 'unit']).agg(value=('value', 'mean')).reset_index()
    # Convert units


def calculate_baseflow(df, method = 'Boughton'):
    dfs = [df]
    for station_id in df['station_id'].unique():
        df_station = df.query(f'constituent == "Q" & station_id == "{station_id}"')[['datetime', 'value']].copy().set_index('datetime')
        if df_station.empty:
            continue
        else:
            df_baseflow = bf.single(df_station['value'], area = None, method = method,return_kge = False)[0][method]
            
            df_baseflow = pd.DataFrame(
                {
                    "station_id": station_id,
                    "station_origin": 'wiski',
                    "datetime": df_baseflow.index,
                    "value": df_baseflow.values,
                    "constituent": 'QB',
                    "unit": 'cfs',
                }
            )
            dfs.append(df_baseflow)
    
    return pd.concat(dfs)


def normalize(df):
    '''
    Standardize raw WISKI data into standardized format without transformations.
    The standardized format includes normalized column names and units.
    ---
    Parameters:
    df (pandas.DataFrame): Raw WISKI data
    Returns:
    pandas.DataFrame: Normalized WISKI data
    '''

    df = convert_units(df)
    df = normalize_columns(df)
    return df

def transform(df, filter_qc_codes = True, data_codes = None, baseflow_method = 'Boughton'):
    '''
    Transform normalized WISKI data into standardized format
    '''
    df = normalize(df)
    if filter_qc_codes:
        if data_codes is None:
            data_codes = DATA_CODES
        df = filter_quality_codes(df, data_codes)
    df = average_results(df)
    df = calculate_baseflow(df, method = baseflow_method)
    df['station_origin'] = 'wiski'
    #df.set_index('datetime',inplace=True)
    return df






#%%
from mpcaHydro.data_manager import dataManager
from pyhcal.repository import Repository
from mpcaHydro import outlets
import duckdb
from mpcaHydro import equis, warehouse, wiski
from hspf.hspfModel import hspfModel
from hspf.uci import UCI



#%% 
'''
New approach. Directly load to warehouse from downloads. 
Store raw and processed data in warehouse. For large timeseries I could store
as parquet files. The transformations using pandas take a bit of time. I imagine doing them
within duckdb would be faster.

'''

# with warehouse.connect(db_path) as con:
#    df = con.execute("SELECT * FROM staging.wiski").df()
#    df = wiski.transform(df,filter_qc_codes = False)

#%%
model_name = 'Nemadji'
db_path = f'C:/Users/mfratki/Documents/{model_name}.duckdb'
start_year = 1996
end_year = 2030
replace = True
filter_qc_codes = True
equis_stations = outlets.equis_stations(model_name)
wiski_stations = outlets.wiski_stations(model_name)
equis.connect('MFRATKI',password = 'DeltaT#MPCA3')
warehouse.init_db(db_path,reset = True)

#%%

with warehouse.connect(db_path) as con:
  df = equis.download(equis_stations)
  warehouse.load_df_to_staging(con,df, 'equis_raw',replace = replace)
  warehouse.load_df_to_analytics(con,equis.transform(df),'equis')

  df = wiski.download(wiski_stations,start_year = start_year, end_year = end_year)
  warehouse.load_df_to_staging(con,df, 'wiski_raw', replace = replace)
  warehouse.load_df_to_analytics(con,wiski.transform(df,filter_qc_codes = filter_qc_codes),'wiski') # method includes normalization

  outlets.build_outlets(con, model_name)


with warehouse.connect(db_path) as con:
  warehouse.update_views(con)


#%%

import requests
url = 'http://ifrshiny.seas.umich.edu/mglp/'
requests.get(url)


 
db_path = 'C:/Users/mfratki/Documents/Rum.duckdb'
modl_db.build_outlet_db(db_path)
con = duckdb.connect(db_path)
con.execute("SELECT * FROM station_reach_pairs").df()
con.execute('SELECT * FROM station_reach_pairs WHERE outlet_id = 76').df()

# Need to remove duplicates from MODL_DB
modl_db.MODL_DB.loc[modl_db.MODL_DB.duplicated(['station_id','source'])]

#%%
dm = dataManager('C:/Users/mfratki/Documents/')
dm._build_warehouse()
equis_stations = modl_db.equis_stations('Nemadji')
wiski_stations = modl_db.wiski_stations('Nemadji')

#%% Old approach. Store as indvidual processed station files then load to warehouse
for station_id in equis_stations:
    dm._download_station_data(station_id,'equis', True)
    
for station_id in wiski_stations:
    dm._download_station_data(station_id,'wiski', True)










#%% Adding HSPF outputs to warehouse











con = duckdb.connect(db_path)

model_name = 'Nemadji'
outlets = [group for _, group in modl_db.MODL_DB.query('repository_name == @model_name').groupby(by = ['opnids','repository_name'])]

for outlet in outlets:
    1+1


dfs = []
for constituent in ['Q','TSS','TP','N','OP','TKN']:
    opnids = modl_db.split_opnids([opnid.split(',') for opnid in set(outlet['opnids'].tolist())])
    for opnid in opnids:
        df = mod.hbns.get_reach_constituent(constituent,opnids,time_step='h')
        df.columns = ['value']
        df['constituent'] = constituent
        df['operation'] = operation
        df['opnid'] = opnid
        dfs.append(df)
        
df = pd.concat(dfs).reset_index()
df['model_name'] = model_name



station_ids = ['H05018001','S006-214','S015-102']
target_constituent = 'TSS'
flow_constituent = 'Q'

# build placeholders for the IN list (one ? per station id)
placeholders = ','.join(['?'] * len(station_ids))

sql = f'''
SELECT o.*, f.datetime AS flow_datetime, f.value AS flow, f.baseflow, f.station_id AS flow_station_id, f.station_origin AS flow_station_origin  
FROM analytics.observations o
JOIN analytics.observations f
  ON o.datetime = f.datetime
WHERE o.constituent = ?
  AND o.station_id IN ({placeholders})
  AND f.constituent = ?;
'''

# parameter order must match the ? positions in the query
params = [target_constituent] + station_ids + [flow_constituent]

df = con.execute(sql, params).df()

outlet_id: station_ids

outlet_id: opnid


outlets = []
for index, (_, group) in enumerate(modl_db.MODL_DB.groupby(by = ['opnids','repository_name'])):
    group['outlet_id'] = index
    group.reset_index(drop=True, inplace=True)
    outlets.append(group)


    for _, row in group.iterrows():
        opnids = group.split_opnids(row['opnids'].str.split(',').to_list())
        row*len(opnids)



# -*- coding: utf-8 -*-
"""
Created on Thu May  1 09:51:51 2025

@author: mfratki
"""
#import sqlite3
from pathlib import Path
import geopandas as gpd
import pandas as pd
import duckdb
#from hspf_tools.calibrator import etlWISKI, etlSWD


#stations_wiski = gpd.read_file('C:/Users/mfratki/Documents/GitHub/pyhcal/src/pyhcal/data/stations_wiski.gpkg')


_stations_wiski = gpd.read_file(str(Path(__file__).resolve().parent/'data\\stations_wiski.gpkg'))
stations_wiski = _stations_wiski.dropna(subset='opnids')[['station_id','true_opnid','opnids','comments','modeled','repository_name','wplmn_flag']]
stations_wiski['source'] = 'wiski'
_stations_equis = gpd.read_file(str(Path(__file__).resolve().parent/'data\\stations_EQUIS.gpkg'))
stations_equis = _stations_equis.dropna(subset='opnids')[['station_id','true_opnid','opnids','comments','modeled','repository_name']]
stations_equis['source'] = 'equis'
stations_equis['wplmn_flag'] = 0


DB_PATH = str(Path(__file__).resolve().parent/'data\\outlets.duckdb')

MODL_DB = pd.concat([stations_wiski,stations_equis])
MODL_DB['opnids'] = MODL_DB['opnids'].str.strip().replace('',pd.NA)
MODL_DB = MODL_DB.dropna(subset='opnids')
MODL_DB = MODL_DB.drop_duplicates(['station_id','source']).reset_index(drop=True)

def _reload():
    global _stations_wiski, stations_wiski, _stations_equis, stations_equis, MODL_DB
    _stations_wiski = gpd.read_file(str(Path(__file__).resolve().parent/'data\\stations_wiski.gpkg'))
    stations_wiski = _stations_wiski.dropna(subset='opnids')[['station_id','true_opnid','opnids','comments','modeled','repository_name','wplmn_flag']]
    stations_wiski['source'] = 'wiski'
    _stations_equis = gpd.read_file(str(Path(__file__).resolve().parent/'data\\stations_EQUIS.gpkg'))
    stations_equis = _stations_equis.dropna(subset='opnids')[['station_id','true_opnid','opnids','comments','modeled','repository_name']]
    stations_equis['source'] = 'equis'
    stations_equis['wplmn_flag'] = 0

    MODL_DB = pd.concat([stations_wiski,stations_equis])
    MODL_DB['opnids'] = MODL_DB['opnids'].str.strip().replace('',pd.NA)
    MODL_DB = MODL_DB.dropna(subset='opnids')
    MODL_DB = MODL_DB.drop_duplicates(['station_id','source']).reset_index(drop=True)


def get_model_db(model_name: str):
    return MODL_DB.query('repository_name == @model_name')

def split_opnids(opnids: list):
    return [abs(int(float(j))) for i in opnids for j in i]

def valid_models():
    return MODL_DB['repository_name'].unique().tolist()

def wplmn_station_opnids(model_name):
    opnids = MODL_DB.query('repository_name == @model_name and wplmn_flag == 1 and source == "wiski"')['opnids'].str.split(',').to_list()
    return split_opnids(opnids)

def wiski_station_opnids(model_name):
    opnids = MODL_DB.query('repository_name == @model_name and source == "wiski"')['opnids'].str.split(',').to_list()
    return split_opnids(opnids)

def equis_station_opnids(model_name):
    opnids = MODL_DB.query('repository_name == @model_name and source == "equis"')['opnids'].str.split(',').to_list()
    return split_opnids(opnids)

def station_opnids(model_name):
    opnids = MODL_DB.query('repository_name == @model_name')['opnids'].str.split(',').to_list()
    return split_opnids(opnids)

def equis_stations(model_name):
    return MODL_DB.query('repository_name == @model_name and source == "equis"')['station_id'].tolist()

def wiski_stations(model_name):
    return MODL_DB.query('repository_name == @model_name and source == "wiski"')['station_id'].tolist()

def wplmn_stations(model_name):
    return MODL_DB.query('repository_name == @model_name and wplmn_flag == 1 and source == "wiski"')['station_id'].tolist()

def outlets(model_name):
    return [group for _, group in MODL_DB.query('repository_name == @model_name').groupby(by = ['opnids','repository_name'])]

def outlet_stations(model_name):
    return [group['station_id'].to_list() for _, group in MODL_DB.query('repository_name == @model_name').groupby(by = ['opnids','repository_name'])]

def _split_opnids(opnids: list):
    return [int(float(j)) for i in opnids for j in i]

def connect(db_path, read_only=True):
    Path(db_path).parent.mkdir(parents=True, exist_ok=True)
    return duckdb.connect(db_path,read_only=read_only)


def init_db(db_path: str,reset: bool = False):
    """
    Initialize the DuckDB database: create staging and analytics schemas
    """
    db_path = Path(db_path)
    if reset and db_path.exists():
        db_path.unlink()

    with connect(db_path.as_posix()) as con:
        con.execute(OUTLETS_SCHEMA)



# Accessors:
def get_outlets_by_model(model_name: str):
    with connect(DB_PATH) as con:
        df = con.execute(
            """
            SELECT r.*
            FROM outlets.station_reach_pairs r
            WHERE r.repository_name = ?
            """,
            [model_name]
        ).fetchdf()
    return df

def get_outlets_by_reach(reach_id: int, model_name: str):
    """
    Return all outlet rows for outlets that include the given reach_id in the given model_name.
    """
    with connect(DB_PATH) as con:
        df = con.execute(
            """
            SELECT r.*
            FROM outlets.station_reach_pairs r
            WHERE r.reach_id = ? AND r.repository_name = ?
            """,
        [reach_id, model_name]).fetchdf()
    return df
 
def get_outlets_by_station(station_id: str, station_origin: str):
    """
    Return all outlet rows for outlets that include the given reach_id in the given model_name.
    """
    with connect(DB_PATH) as con:

        df = con.execute(
        """
        SELECT r.*
        FROM outlets.station_reach_pairs r
        WHERE r.station_id = ? AND r.station_origin = ?
        """,
        [station_id, station_origin]).fetchdf()
    return df



class OutletGateway:
    def __init__(self, model_name: str):
        self.model_name = model_name
        self.db_path = DB_PATH
        self.modl_db = get_model_db(model_name)

    # Legacy methods to access functions
    def wplmn_station_opnids(self):
        return wplmn_station_opnids(self.model_name)

    def wiski_station_opnids(self):
        return wiski_station_opnids(self.model_name)

    def equis_station_opnids(self):
        return equis_station_opnids(self.model_name)

    def station_opnids(self):
        return station_opnids(self.model_name)

    def equis_stations(self):
        return equis_stations(self.model_name)

    def wiski_stations(self):
        return wiski_stations(self.model_name)

    def wplmn_stations(self):
        return wplmn_stations(self.model_name)

    def outlets(self):
        return outlets(self.model_name)

    def outlet_stations(self):
        return outlet_stations(self.model_name)

    # Accessors for outlets
    def get_outlets(self):
        return get_outlets_by_model(self.model_name)

    def get_outlets_by_reach(self, reach_id: int):
        return get_outlets_by_reach(reach_id, self.model_name)

    def get_outlets_by_station(self, station_id: str, station_origin: str):
        assert(station_id in self.wiski_stations() + self.equis_stations()), f"Station ID {station_id} not found in model {self.model_name}"
        return get_outlets_by_station(station_id, station_origin)

# constructors:
def build_outlet_db(db_path: str = None):
    if db_path is None:
        db_path = DB_PATH
    init_db(db_path,reset=True)
    with connect(db_path) as con:
        build_outlets(con)


def build_outlets(con, model_name: str = None):
    if model_name is not None:
        modl_db = get_model_db(model_name)
    else:
        modl_db = MODL_DB

    for index, (_, group) in enumerate(modl_db.drop_duplicates(['station_id','source']).groupby(by = ['opnids','repository_name'])):
        repo_name = group['repository_name'].iloc[0]    
        add_outlet(con, outlet_id = index, outlet_name = None, repository_name = repo_name, notes = None)
        
        opnids = set(_split_opnids(group['opnids'].str.split(',').to_list()))

        for opnid in opnids:
            if opnid < 0:
                exclude = 1
            else:
                exclude = 0
            add_reach(con, outlet_id = index, reach_id = abs(opnid),exclude = exclude, repository_name = repo_name)

        for _, row in group.drop_duplicates(subset=['station_id', 'source']).iterrows():
            add_station(con, outlet_id = index, station_id = row['station_id'], station_origin = row['source'], true_opnid = row['true_opnid'], repository_name= repo_name, comments = row['comments'])

 
def create_outlet_schema(con, model_name : str):
    for index, (_, group) in enumerate(modl_db.outlets(model_name)):
        repo_name = group['repository_name'].iloc[0]    
        add_outlet(con, outlet_id = index, outlet_name = None, repository_name = repo_name, notes = None)
        
        opnids = set(_split_opnids(group['opnids'].str.split(',').to_list()))

        for opnid in opnids:
            if opnid < 0:
                exclude = 1
            else:
                exclude = 0
            add_reach(con, outlet_id = index, reach_id = abs(opnid),exclude = exclude, repository_name = repo_name)

        for _, row in group.drop_duplicates(subset=['station_id', 'source']).iterrows():
            add_station(con, outlet_id = index, station_id = row['station_id'], station_origin = row['source'], true_opnid = row['true_opnid'], repository_name= repo_name, comments = row['comments'])


def add_outlet(con,
               outlet_id: str,
               repository_name: str,
               outlet_name = None,
               notes = None):
    """
    Insert an outlet. repository_name is required.
    """
    con.execute(
        "INSERT INTO outlets.outlets (outlet_id, repository_name, outlet_name, notes) VALUES (?, ?, ?, ?)",
        [outlet_id, repository_name, outlet_name, notes]
    )

def add_station(con,
                outlet_id: str,
                station_id: str,
                station_origin: str,
                true_opnid: str,
                repository_name: str,
                comments = None):
    """
    Insert a station membership for an outlet.
    Constraints:
    - PRIMARY KEY (station_id, station_origin): unique per origin across all outlets.
    - true_opnid and true_opnid_repository_name are required per schema.
    """
    con.execute(
        """INSERT INTO outlets.outlet_stations
           (outlet_id, station_id, station_origin, true_opnid, repository_name, comments)
           VALUES (?, ?, ?, ?, ?, ?)""",
        [outlet_id, station_id, station_origin, true_opnid, repository_name, comments]
    )

def add_reach(con,
              outlet_id: str,
              reach_id: str,
              repository_name: str,
              exclude: int = 0):
    """
    Insert a reach membership for an outlet.
    - repository_name is required and participates in the PK (reach_id, repository_name).
    - exclude = 1 to mark a reach as excluded from association views.
    """
    con.execute(
        """INSERT INTO outlets.outlet_reaches (outlet_id, reach_id, repository_name, exclude)
           VALUES (?, ?, ?, ?)""",
        [outlet_id, reach_id, repository_name, int(exclude)]
    )


OUTLETS_SCHEMA  = """-- schema.sql
-- Simple 3-table design to manage associations between model reaches and observation stations via outlets.
-- Compatible with DuckDB and SQLite.

-- Table 1: outlets
-- Represents a logical grouping that ties stations and reaches together.
CREATE TABLE IF NOT EXISTS outlets (
  outlet_id TEXT PRIMARY KEY,
  repository_name TEXT NOT NULL,
  outlet_name TEXT,
  notes TEXT             -- optional: general notes about the outlet grouping
);

-- Table 2: outlet_stations
-- One-to-many: outlet -> stations
CREATE TABLE IF NOT EXISTS outlet_stations (
  outlet_id TEXT NOT NULL,
  station_id TEXT NOT NULL,
  station_origin TEXT NOT NULL,       -- e.g., 'wiski', 'equis'
  repository_name TEXT NOT NULL,  -- repository model the station is physically located in
  true_opnid TEXT NOT NULL,           -- The specific reach the station physically sits on (optional)
  comments TEXT,             -- Per-station comments, issues, etc.
  CONSTRAINT uq_station_origin UNIQUE (station_id, station_origin),
  FOREIGN KEY (outlet_id) REFERENCES outlets(outlet_id)
);

-- Table 3: outlet_reaches
-- One-to-many: outlet -> reaches
-- A reach can appear in multiple outlets, enabling many-to-many overall.
CREATE TABLE IF NOT EXISTS outlet_reaches (
  outlet_id TEXT NOT NULL,
  reach_id TEXT NOT NULL,    -- model reach identifier (aka opind)
  repository_name TEXT NOT NULL,  -- optional: where the mapping comes from
  exclude INTEGER DEFAULT 0, -- flag to indicate if this reach should be excluded (1) or included (0)
  FOREIGN KEY (outlet_id) REFERENCES outlets(outlet_id)
);

-- Useful views:

-- View: station_reach_pairs
-- Derives the implicit many-to-many station <-> reach relationship via shared outlet_id
CREATE VIEW IF NOT EXISTS station_reach_pairs AS
SELECT
  s.outlet_id,
  s.station_id,
  s.station_origin,
  r.reach_id,
  r.exclude,
  r.repository_name,
FROM outlet_stations s
JOIN outlet_reaches r
  ON s.outlet_id = r.outlet_id;

"""  
    
#row = modl_db.MODL_DB.iloc[0]

#info = etlWISKI.info(row['station_id'])

#modl_db.MODL_DB.query('source == "equis"')

# outlet_dict = {'stations': {'wiski': ['E66050001'],
#                'equis': ['S002-118']},
#                'reaches': {'Clearwater': [650]}
                      



# station_ids = ['S002-118']
# #station_ids = ['E66050001']
# reach_ids = [650]
# flow_station_ids =  ['E66050001']



from pathlib import Path
import duckdb
import glob

#TODO ensure all reports are actually in the reports schema

class reportManager():
    def __init__(self,db_path:Path):
        self.db_path = db_path

    def wiski_qc_counts(self):
        with duckdb.connect(self.db_path,read_only=True) as con:
            return wiski_qc_counts(con)
        
    def constituent_summary(self,constituent: str = None):
        with duckdb.connect(self.db_path,read_only=True) as con:
            return constituent_summary(con,constituent)
        
    def station_reach_pairs(self):
        with duckdb.connect(self.db_path,read_only=True) as con:
            return station_reach_pairs(con)
        
    def outlet_summary(self):
        with duckdb.connect(self.db_path,read_only=True) as con:
            return outlet_summary(con)
        


def outlet_summary(con: duckdb.DuckDBPyConnection):
    query = '''
    SELECT *,
    FROM 
        reports.outlet_constituent_summary
    ORDER BY
        outlet_id,
        constituent
    '''
    df = con.execute(query).fetch_df()
    return df
        

def wiski_qc_counts(con: duckdb.DuckDBPyConnection):
    query = '''
    SELECT *,
    FROM 
        staging.wiski_qc_count
    ORDER BY
        station_no,
        parametertype_name
    '''
    df = con.execute(query).fetch_df()
    return df

def constituent_summary(con: duckdb.DuckDBPyConnection,constituent: str = None):
    
    query = '''
    SELECT *,
    FROM 
        reports.constituent_summary
    ORDER BY
        station_id,
        station_origin,
        constituent
    '''
    df = con.execute(query).fetch_df()
    if constituent is not None:
        df = df[df['constituent'] == constituent]
    return df

def station_reach_pairs(con: duckdb.DuckDBPyConnection):
    query = '''
    SELECT *,
    FROM 
        reports.station_reach_pairs
    ORDER BY
        outlet_id,
        station_id
    '''
    df = con.execute(query).fetch_df()
    return df


# -*- coding: utf-8 -*-
"""
Created on Tue Oct  3 08:04:49 2023

@author: mfratki
"""



import pandas as pd
from mpcaHydro.pyWISK import pyWISK
import time


PARAMETERS_MAP={'5004':'TP Load',
                '5005':'TP Conc',
                '5014':'TSS Load',
                '5015':'TSS Conc',
                '5024':'N Load',
                '5025':'N Conc',
                '5034':'OP Load',
                '5035':'OP Conc',
                '5044':'TKN Load',
                '5045':'TKN Conc'}#,
                #'262' :'Flow'}

def download(station_nos):
    
   
    wiski = pyWISK()
    ts_ids = wiski.get_ts_ids(station_nos = station_nos,
                      stationgroup_id = '1319204',
                      stationparameter_no = list(PARAMETERS_MAP.keys()),
                      ts_name = ['20.Day.Mean'])
    
    flow_ts_ids = wiski.get_ts_ids(station_nos = station_nos,
                      stationgroup_id = '1319204',
                      stationparameter_no = '262',
                      ts_name = ['20.Day.Mean.Archive'])
    
    ts_ids = pd.concat([ts_ids,flow_ts_ids])
    
    if len(ts_ids) == 0:
        print('No WPLMN Sites Available')
        return pd.DataFrame() 
    
    dfs = []
    for ts_id in ts_ids['ts_id']:
        dfs.append(wiski.get_ts(ts_id))
        time.sleep(1)
    
    return pd.concat(dfs)

def transform(data):
    # Constants related to the WPLMN data in the Kisters database
    #UNITS = {'Conc':'5','Mass':'4'}
    #PARAMETER_NOS = ['5004','5005','5014','5015','5024','5025','5034','5035','5044','5045','262']
    NO_DATA_CODES = [255,151,-1]
    #MODELED_CODES = [34,2,30,31,32,70,140,80,15,10,20,27,180,35,141]
    #MEASURED_CODES = [3]
    
    data['datetime'] = pd.to_datetime(data['Timestamp'])
    #data['date'] = data['datetime'].dt.date
    #data['time'] = pd.NA
    data = data.loc[~data['Quality Code'].isin(NO_DATA_CODES),:]
    data.rename(columns = {'Value': 'value',
                           'stationparameter_name': 'variable',
                           'station_name': 'station_name',
                           'station_no': 'station_id',
                           'Quality Code': 'quality_id',
                           'ts_unitsymbol':'unit'},inplace = True)
    data['constituent'] = data['stationparameter_no'].map({'5005': 'TP',
                                                           '5004': 'TP',
                                                           '5014': 'TSS',
                                                           '5015': 'TSS',
                                                           '5024': 'N',
                                                           '5025': 'N',
                                                           '5034': 'OP',
                                                           '5035': 'OP',
                                                           '5044': 'TKN',
                                                           '5045': 'TKN',
                                                           '262' : 'Q'})
    data = data.loc[:,['datetime','value','variable','unit','quality_id','station_id','station_name','constituent']]
    
    data['unit'].replace('ft³/s','cfs',inplace=True)
    data['station_origin'] = 'wplmn'
    return data
    
    # data['Timestamp'] = pd.to_datetime(data['Timestamp']).dt.date
    # data.set_index('Timestamp',drop=True,inplace = True)
    # data['Qc Desc'] = 'modeled point'
    # data.loc[data['Quality Code'] == 3,'Qc Desc'] = 'discrete sample point'
    # data.loc[data['Quality Code'].isin(NO_DATA_CODES),'Qc Desc'] = 'missing data'
    # data.rename(columns = {'Timestamp': 'datetime',
    #                             'Value': 'value',
    #                             'stationparameter_name': 'variable',
    #                             'station_name': 'station_name',
    #                             'station_no': 'station_id'},inplace = True)
    # data['Var Des'] = data['stationparameter_no'].map(PARAMETERS_MAP)

    return data

def load(data,file_path):
    data.to_csv(file_path)